import torch
from torch.optim import SGD, Adam, AdamW


def get_optimizer(hyperparams: dict, model: torch.nn.Module) -> torch.optim.Optimizer:

    optimizer: torch.optim.Optimizer

    if hyperparams['optimizer'] == "adam":
        optimizer = Adam(
            model.parameters(), 
            lr=hyperparams['lr'],
            betas=(0.9, 0.999),
            weight_decay=hyperparams['weight_decay']
        )
    elif hyperparams['optimizer'] == "adamw":
        optimizer = AdamW(
            model.parameters(), 
            lr=hyperparams['lr'],
            betas=(0.9, 0.999),
            weight_decay=hyperparams['weight_decay']
        )
    elif hyperparams['optimizer'] == "sgd":
        optimizer = SGD(
            model.parameters(),
            lr=hyperparams['lr'],
            weight_decay=hyperparams['weight_decay']
        )
    else:
        raise ValueError("Optimizer not recognized")
    
    return optimizer



import torch
from torch.optim import SGD, Adam, AdamW

from ..._internal.config.config_pretrain import ConfigPretrain


def get_optimizer(hyperparams: dict, model: torch.nn.Module) -> torch.optim.Optimizer:

    optimizer: torch.optim.Optimizer

    if hyperparams['optimizer'] == "adam":
        optimizer = Adam(
            model.parameters(), 
            lr=hyperparams['lr'],
            betas=(0.9, 0.999),
            weight_decay=hyperparams['weight_decay']
        )
    elif hyperparams['optimizer'] == "adamw":
        optimizer = AdamW(
            model.parameters(), 
            lr=hyperparams['lr'],
            betas=(0.9, 0.999),
            weight_decay=hyperparams['weight_decay']
        )
    elif hyperparams['optimizer'] == "sgd":
        optimizer = SGD(
            model.parameters(),
            lr=hyperparams['lr'],
            weight_decay=hyperparams['weight_decay']
        )
    else:
        raise ValueError("Optimizer not recognized")
    
    return optimizer


def get_optimizer_pretrain(cfg: ConfigPretrain, model: torch.nn.Module) -> torch.optim.Optimizer:

    parameters = [(name, param) for name, param in model.named_parameters()]

    parameters_with_weight_decay = []
    parameters_without_weight_decay = []

    for name, param in parameters:
        if name.endswith("bias") or 'norm' in name or 'embedding' in name:
            parameters_without_weight_decay.append(param)
        else:
            parameters_with_weight_decay.append(param)

    optimizer_parameters = [
        {"params": parameters_with_weight_decay, "weight_decay": cfg.optim.weight_decay},
        {"params": parameters_without_weight_decay, "weight_decay": 0.0},
    ]

    optimizer = torch.optim.AdamW(
        optimizer_parameters, 
        lr=cfg.optim.lr,
        betas=(cfg.optim.beta1, cfg.optim.beta2),
        weight_decay=cfg.optim.weight_decay
    )
    
    return optimizer


class GradScaler(torch.amp.GradScaler):

    def __init__(
        self, 
        enabled: bool = True,
        scale_init: float = 2.**16,
        scale_min: float = 1.,
        growth_interval: int = 2000,
        device: str = 'cuda'
    ):
        super().__init__(enabled=enabled, device="cpu", init_scale=scale_init, growth_interval=growth_interval) # type: ignore
        self._enabled = enabled
        self.scale_min = scale_min
        self.device = device

        if not self._enabled:
            # We write scale=1 to log if the scaler is disabled
            self._scale = torch.tensor((1,), dtype=torch.float32, device=self.device)


    def update(self):

        if not self._enabled:
            return

        super().update()

        if self._scale < self.scale_min:
            super().update(self.scale_min)


def move_optimizer_to(optim, device):
    for param in optim.state.values():
        # Not sure there are any global tensors in the state dict
        if isinstance(param, torch.Tensor):
            param.data = param.data.to(device)
            if param._grad is not None:
                param._grad.data = param._grad.data.to(device)
        elif isinstance(param, dict):
            for subparam in param.values():
                if isinstance(subparam, torch.Tensor):
                    subparam.data = subparam.data.to(device)
                    if subparam._grad is not None:
                        subparam._grad.data = subparam._grad.data.to(device)


# The original implementation in this file was based on scikit-garden that comes under the following license.
# The current version of the code has been modified beyond its original version.

# New BSD License

# Copyright (c) 2016 - scikit-garden developers.

# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#   c. Neither the name of the scikit-garden developers nor the names of
#      its contributors may be used to endorse or promote products
#      derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

import logging
from functools import partial

import numpy as np
import pandas as pd
from sklearn.ensemble._forest import ForestRegressor
from sklearn.tree import BaseDecisionTree, DecisionTreeRegressor, ExtraTreeRegressor
from sklearn.utils import check_array, check_random_state, check_X_y

logger = logging.getLogger(__name__)


def weighted_percentile(a, q, weights=None, sorter=None, is_filtered=False):
    """
    Returns the weighted percentile of a at q given weights.

    Parameters
    ----------
    a: array-like, shape=(n_samples,)
        samples at which the quantile.
    q: int
        quantile between 0 and 100.
    weights: array-like, shape=(n_samples,)
        weights[i] is the weight given to point a[i] while computing the
        quantile. If weights[i] is zero, a[i] is simply ignored during the
        percentile computation.
    sorter: array-like, shape=(n_samples,)
        If provided, assume that a[sorter] is sorted.
    is_filtered: bool
        If True, weights is assumed to contain only non-zero values.

    Returns
    -------
    percentile: float
        Weighted percentile of a at q.

    References
    ----------
    1. https://en.wikipedia.org/wiki/Percentile#The_Weighted_Percentile_method

    Notes
    -----
    Note that weighted_percentile(a, q) is not equivalent to
    np.percentile(a, q). This is because in np.percentile
    sorted(a)[i] is assumed to be at quantile 0.0, while here we assume
    sorted(a)[i] is given a weight of 1.0 / len(a), hence it is at the
    1.0 / len(a)th quantile.
    """
    if weights is None:
        weights = np.ones_like(a)
    if q > 100 or q < 0:
        raise ValueError("q should be in-between 0 and 100, " "got %d" % q)

    a = np.asarray(a, dtype=np.float32)
    weights = np.asarray(weights, dtype=np.float32)
    if len(a) != len(weights):
        raise ValueError("a and weights should have the same length.")

    if sorter is not None:
        a = a[sorter]
        weights = weights[sorter]

    if not is_filtered:
        nz = weights != 0
        a = a[nz]
        weights = weights[nz]

    if sorter is None:
        sorted_indices = np.argsort(a)
        sorted_a = a[sorted_indices]
        sorted_weights = weights[sorted_indices]
    else:
        sorted_a = a
        sorted_weights = weights

    # Step 1
    sorted_cum_weights = np.cumsum(sorted_weights)
    total = sorted_cum_weights[-1]

    # Step 2
    partial_sum = 100.0 / total * (sorted_cum_weights - sorted_weights / 2.0)
    start = np.searchsorted(partial_sum, q) - 1
    if start == len(sorted_cum_weights) - 1:
        return sorted_a[-1]
    if start == -1:
        return sorted_a[0]

    # Step 3.
    fraction = (q - partial_sum[start]) / (partial_sum[start + 1] - partial_sum[start])
    return sorted_a[start] + fraction * (sorted_a[start + 1] - sorted_a[start])


class BaseTreeQuantileRegressor(BaseDecisionTree):
    def predict(self, X, quantile=None, check_input=False):
        """
        Predict regression value for X.

        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_samples, n_features]
            The input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csr_matrix``.

        quantile : int, optional
            Value ranging from 0 to 100. By default, the mean is returned.

        check_input : boolean, (default=True)
            Allow to bypass several input checking.
            Don't use this parameter unless you know what you do.

        Returns
        -------
        y : array of shape = [n_samples]
            If quantile is set to None, then return E(Y | X). Else return
            y such that F(Y=y | x) = quantile.
        """
        # apply method requires X to be of dtype np.float32
        X = check_array(X, dtype=np.float32, accept_sparse="csc")
        if quantile is None:
            return super(BaseTreeQuantileRegressor, self).predict(X, check_input=check_input)

        quantiles = np.zeros(X.shape[0])
        X_leaves = self.apply(X)
        unique_leaves = np.unique(X_leaves)
        for leaf in unique_leaves:
            quantiles[X_leaves == leaf] = weighted_percentile(self.y_train_[self.y_train_leaves_ == leaf], quantile)
        return quantiles

    def fit(self, X, y, sample_weight=None, check_input=True):
        """
        Build a decision tree classifier from the training set (X, y).

        Parameters
        ----------
        X : array-like or sparse matrix, shape = [n_samples, n_features]
            The training input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csc_matrix``.

        y : array-like, shape = [n_samples] or [n_samples, n_outputs]
            The target values (class labels) as integers or strings.

        sample_weight : array-like, shape = [n_samples] or None
            Sample weights. If None, then samples are equally weighted. Splits
            that would create child nodes with net zero or negative weight are
            ignored while searching for a split in each node. Splits are also
            ignored if they would result in any single class carrying a
            negative weight in either child node.

        check_input : boolean, (default=True)
            Allow to bypass several input checking.
            Don't use this parameter unless you know what you do.

        Returns
        -------
        self : object
            Returns self.
        """
        # y passed from a forest is 2-D. This is to silence the
        # annoying data-conversion warnings.
        y = np.asarray(y)
        if np.ndim(y) == 2 and y.shape[1] == 1:
            y = np.ravel(y)

        # apply method requires X to be of dtype np.float32
        X, y = check_X_y(X, y, accept_sparse="csc", dtype=np.float32, multi_output=False)
        super(BaseTreeQuantileRegressor, self).fit(X, y, sample_weight=sample_weight, check_input=check_input)
        self.y_train_ = y

        # Stores the leaf nodes that the samples lie in.
        self.y_train_leaves_ = self.tree_.apply(X)
        return self


class DecisionTreeQuantileRegressor(BaseTreeQuantileRegressor, DecisionTreeRegressor):
    """A decision tree regressor that provides quantile estimates.

    Parameters
    ----------
    criterion : string, optional (default="squared_error")
        The function to measure the quality of a split. Supported criteria
        are "squared_error" for the mean squared error, which is equal to variance
        reduction as feature selection criterion, and "mae" for the mean
        absolute error.
        .. versionadded:: 0.18
           Mean Absolute Error (MAE) criterion.

    splitter : string, optional (default="best")
        The strategy used to choose the split at each node. Supported
        strategies are "best" to choose the best split and "random" to choose
        the best random split.

    max_features : int, float, string or None, optional (default=None)
        The number of features to consider when looking for the best split:
        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a percentage and
          `int(max_features * n_features)` features are considered at each
          split.
        - If "auto", then `max_features=n_features`.
        - If "sqrt", then `max_features=sqrt(n_features)`.
        - If "log2", then `max_features=log2(n_features)`.
        - If None, then `max_features=n_features`.
        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

    max_depth : int or None, optional (default=None)
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples.

    min_samples_split : int, float, optional (default=2)
        The minimum number of samples required to split an internal node:
        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a percentage and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.
        .. versionchanged:: 0.18
           Added float values for percentages.

    min_samples_leaf : int, float, optional (default=1)
        The minimum number of samples required to be at a leaf node:
        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a percentage and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.
        .. versionchanged:: 0.18
           Added float values for percentages.

    max_leaf_nodes : int or None, optional (default=None)
        Grow a tree with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    presort : bool, optional (default=False)
        Whether to presort the data to speed up the finding of best splits in
        fitting. For the default settings of a decision tree on large
        datasets, setting this to true may slow down the training process.
        When using either a smaller dataset or a restricted depth, this may
        speed up the training.

    Attributes
    ----------
    feature_importances_ : array of shape = [n_features]
        The feature importances.
        The higher, the more important the feature.
        The importance of a feature is computed as the
        (normalized) total reduction of the criterion brought
        by that feature. It is also known as the Gini importance [4]_.

    max_features_ : int,
        The inferred value of max_features.

    n_features_ : int
        The number of features when ``fit`` is performed.

    n_outputs_ : int
        The number of outputs when ``fit`` is performed.

    tree_ : Tree object
        The underlying Tree object.

    y_train_ : array-like
        Train target values.

    y_train_leaves_ : array-like.
        Cache the leaf nodes that each training sample falls into.
        y_train_leaves_[i] is the leaf that y_train[i] ends up at.
    """

    def __init__(
        self,
        criterion="squared_error",
        splitter="best",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        max_features=None,
        random_state=None,
        max_leaf_nodes=None,
    ):
        super(DecisionTreeQuantileRegressor, self).__init__(
            criterion=criterion,
            splitter=splitter,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            random_state=random_state,
        )


class ExtraTreeQuantileRegressor(BaseTreeQuantileRegressor, ExtraTreeRegressor):
    def __init__(
        self,
        criterion="squared_error",
        splitter="random",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        max_features=1.0,
        random_state=None,
        max_leaf_nodes=None,
    ):
        super(ExtraTreeQuantileRegressor, self).__init__(
            criterion=criterion,
            splitter=splitter,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            random_state=random_state,
        )


def generate_sample_indices(random_state, n_samples):
    """
    Generates bootstrap indices for each tree fit.

    Parameters
    ----------
    random_state: int, RandomState instance or None
        If int, random_state is the seed used by the random number generator.
        If RandomState instance, random_state is the random number generator.
        If None, the random number generator is the RandomState instance used
        by np.random.
    n_samples: int
        Number of samples to generate from each tree.

    Returns
    -------
    sample_indices: array-like, shape=(n_samples), dtype=np.int32
        Sample indices.
    """
    random_instance = check_random_state(random_state)
    sample_indices = random_instance.randint(0, n_samples, n_samples)
    return sample_indices


def get_weighted_neighbors_dataframe(X_leaves, y_train_leaves, y_train, y_weights):
    """For each test sample, get the list of weighted targets that are assigned to the same leaf by at least one estimator.

    Parameters
    ----------
    X_leaves : array, shape [n_test, n_estimators]
        Index of the leave assigned to each test sample by each estimator.
    y_train_leaves : array, shape [n_estimators, n_train]
        Index of the leave assigned to each training sample by each estimator.
    y_train : array, shape [n_train]
        Values of training samples.
    y_weights : array, shape [n_estimators, n_train]
        Weight assigned to each training sample by each estimator.

    Returns
    -------
    weighted_neighbors_dataframe : pd.DataFrame
        Dataframe that contains weighted neighbors of each item in the test set.
        Columns:
            item_id: ID of each item in the test set
            y: Target value encountered in the same leaf as the item
            weight: Weight assigned to each target value
    """
    assert X_leaves.shape[1] == y_train_leaves.shape[0] == y_weights.shape[0]
    assert y_train_leaves.shape[1] == y_train.shape[0] == y_weights.shape[1]

    num_test, num_trees = X_leaves.shape
    _, num_train = y_train_leaves.shape

    tree_index_x = np.arange(num_trees)[None].repeat([num_test], axis=0)
    item_index_x = np.arange(num_test)[:, None].repeat([num_trees], axis=1)
    df_x = pd.DataFrame(
        {
            "item_id": item_index_x.ravel(),
            "tree_id": tree_index_x.ravel(),
            "leaf": X_leaves.ravel(),
        }
    )
    tree_index_y = np.arange(num_trees)[:, None].repeat([num_train], axis=1)
    target_y = y_train[None].repeat([num_trees], axis=0)
    df_y = pd.DataFrame(
        {
            "tree_id": tree_index_y.ravel(),
            "leaf": y_train_leaves.ravel(),
            "y": target_y.ravel(),
            "weight": y_weights.ravel(),
        }
    )
    samples_with_neighbors = pd.merge(df_x, df_y, on=["tree_id", "leaf"])
    return samples_with_neighbors.groupby(["item_id", "y"]).sum().reset_index()[["item_id", "y", "weight"]]


def get_quantiles(neighbors_df, quantile_levels):
    """Compute predicted quantiles for the given sample.

    Parameters
    ----------
    neighbors_df : pd.DataFrame
        DataFrame with columns y (target values for each sample) and weight (weight assigned to each sample)
    quantile_levels : List[float]
        List of quantiles to predict between 0.0 and 1.0

    Returns
    -------
    quantiles : array, shape [len(quantile_levels)]
        Predicted quantiles.
    """
    result = []
    for q in quantile_levels:
        result.append(
            weighted_percentile(
                neighbors_df.y,
                int(q * 100),
                neighbors_df.weight,
                sorter=slice(None),  # targets are already sorted, so no sorting required
                is_filtered=True,
            )
        )
    return result


class BaseForestQuantileRegressor(ForestRegressor):
    def fit(self, X, y, sample_weight=None):
        """
        Build a forest from the training set (X, y).

        Parameters
        ----------
        X : array-like or sparse matrix, shape = [n_samples, n_features]
            The training input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csc_matrix``.
        y : array-like, shape = [n_samples] or [n_samples, n_outputs]
            The target values (class labels) as integers or strings.
        sample_weight : array-like, shape = [n_samples] or None
            Sample weights. If None, then samples are equally weighted. Splits
            that would create child nodes with net zero or negative weight are
            ignored while searching for a split in each node. Splits are also
            ignored if they would result in any single class carrying a
            negative weight in either child node.

        Returns
        -------
        self : object
            Returns self.
        """
        logger.warning(
            f"\tWARNING: {self.__class__.__name__} are experimental for quantile regression. "
            f"They may change or be removed without warning in future releases."
        )
        if sample_weight is not None:
            logger.warning(f"\tWARNING: {self.__class__.__name__} ignores sample_weight.")

        # apply method requires X to be of dtype np.float32
        X, y = check_X_y(X, y, accept_sparse="csc", dtype=np.float32, multi_output=False)
        super(BaseForestQuantileRegressor, self).fit(X, y)

        self.y_train_ = y
        self.y_train_leaves_ = -np.ones((self.n_estimators, len(y)), dtype=np.int32)
        self.y_weights_ = np.zeros_like((self.y_train_leaves_), dtype=np.float32)

        for i, est in enumerate(self.estimators_):
            if self.bootstrap:
                bootstrap_indices = generate_sample_indices(est.random_state, len(y))
            else:
                bootstrap_indices = np.arange(len(y))

            est_weights = np.bincount(bootstrap_indices, minlength=len(y))
            # FIXME: When updating from scikit-learn 1.3.2 to 1.4.0, BaseTreeQuantileRegressor.fit is not called
            # Re-calculating y_train_ and y_train_leaves_ to resolve this issue
            est.y_train_ = y
            est.y_train_leaves_ = est.tree_.apply(X)
            y_train_leaves = est.y_train_leaves_
            # Normalize the bootstrap weights such that the total weight of each leaf sums up to 1
            # Relabel leaves starting from zero in order to efficiently count the total sum per leaf with bincount
            leaves_starting_from_zero = np.unique(y_train_leaves, return_inverse=True)[1]
            weight_per_leaf = np.bincount(leaves_starting_from_zero, weights=est_weights)
            self.y_weights_[i] = est_weights / weight_per_leaf[leaves_starting_from_zero]

            self.y_train_leaves_[i, bootstrap_indices] = y_train_leaves[bootstrap_indices]
        return self

    def predict(self, X, quantile_levels=None):
        """
        Predict regression value for X.

        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_samples, n_features]
            The input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csr_matrix``.
        quantile_levels : List[float], optional
            List of quantiles (between 0.0 and 1.0) to predict. If not provided, mean is returned.

        Returns
        -------
        y : array
            If quantile_levels is None, then y contains E(Y | X) and has shape [n_samples].
            Otherwise, y contains the predicted quantiles and has shape [n_samples, len(quantile_levels)]
        """
        # apply method requires X to be of dtype np.float32
        X = check_array(X, dtype=np.float32, accept_sparse="csc")
        if quantile_levels is None:
            return super(BaseForestQuantileRegressor, self).predict(X)
        elif isinstance(quantile_levels, float):
            quantile_levels = [quantile_levels]

        X_leaves = self.apply(X)

        samples_with_weighted_neighbors = get_weighted_neighbors_dataframe(
            X_leaves=X_leaves, y_train_leaves=self.y_train_leaves_, y_train=self.y_train_, y_weights=self.y_weights_
        )
        quantile_preds = samples_with_weighted_neighbors.groupby("item_id").apply(partial(get_quantiles, quantile_levels=quantile_levels), include_groups=False)
        return np.stack(quantile_preds.values.tolist())


class RandomForestQuantileRegressor(BaseForestQuantileRegressor):
    """
    A random forest regressor that provides quantile estimates.
    A random forest is a meta estimator that fits a number of classifying
    decision trees on various sub-samples of the dataset and use averaging
    to improve the predictive accuracy and control over-fitting.
    The sub-sample size is always the same as the original
    input sample size but the samples are drawn with replacement if
    `bootstrap=True` (default).

    Parameters
    ----------
    n_estimators : integer, optional (default=10)
        The number of trees in the forest.
    criterion : string, optional (default="squared_error")
        The function to measure the quality of a split. Supported criteria
        are "squared_error" for the mean squared error, which is equal to variance
        reduction as feature selection criterion, and "mae" for the mean
        absolute error.
        .. versionadded:: 0.18
           Mean Absolute Error (MAE) criterion.
    max_features : int, float, string or None, optional (default="auto")
        The number of features to consider when looking for the best split:
        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a percentage and
          `int(max_features * n_features)` features are considered at each
          split.
        - If "auto", then `max_features=n_features`.
        - If "sqrt", then `max_features=sqrt(n_features)`.
        - If "log2", then `max_features=log2(n_features)`.
        - If None, then `max_features=n_features`.
        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.
    max_depth : integer or None, optional (default=None)
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples.
    min_samples_split : int, float, optional (default=2)
        The minimum number of samples required to split an internal node:
        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a percentage and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.
        .. versionchanged:: 0.18
           Added float values for percentages.
    min_samples_leaf : int, float, optional (default=1)
        The minimum number of samples required to be at a leaf node:
        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a percentage and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.
        .. versionchanged:: 0.18
           Added float values for percentages.
    max_leaf_nodes : int or None, optional (default=None)
        Grow trees with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.
    bootstrap : boolean, optional (default=True)
        Whether bootstrap samples are used when building trees.
    oob_score : bool, optional (default=False)
        whether to use out-of-bag samples to estimate
        the R^2 on unseen data.
    n_jobs : integer, optional (default=1)
        The number of jobs to run in parallel for both `fit` and `predict`.
        If -1, then the number of jobs is set to the number of cores.
    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.
    verbose : int, optional (default=0)
        Controls the verbosity of the tree building process.
    warm_start : bool, optional (default=False)
        When set to ``True``, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just fit a whole
        new forest.
    Attributes
    ----------
    estimators_ : list of DecisionTreeQuantileRegressor
        The collection of fitted sub-estimators.
    feature_importances_ : array of shape = [n_features]
        The feature importances (the higher, the more important the feature).
    n_features_ : int
        The number of features when ``fit`` is performed.
    n_outputs_ : int
        The number of outputs when ``fit`` is performed.
    oob_score_ : float
        Score of the training dataset obtained using an out-of-bag estimate.
    oob_prediction_ : array of shape = [n_samples]
        Prediction computed with out-of-bag estimate on the training set.
    y_train_ : array-like, shape=(n_samples,)
        Cache the target values at fit time.
    y_weights_ : array-like, shape=(n_estimators, n_samples)
        y_weights_[i, j] is the weight given to sample ``j` while
        estimator ``i`` is fit. If bootstrap is set to True, this
        reduces to a 2-D array of ones.
    y_train_leaves_ : array-like, shape=(n_estimators, n_samples)
        y_train_leaves_[i, j] provides the leaf node that y_train_[i]
        ends up when estimator j is fit. If y_train_[i] is given
        a weight of zero when estimator j is fit, then the value is -1.

    References
    ----------
    .. [1] Nicolai Meinshausen, Quantile Regression Forests
        http://www.jmlr.org/papers/volume7/meinshausen06a/meinshausen06a.pdf
    """

    def __init__(
        self,
        n_estimators=10,
        criterion="squared_error",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        max_features=1.0,
        max_leaf_nodes=None,
        bootstrap=True,
        oob_score=False,
        n_jobs=1,
        random_state=None,
        verbose=0,
        warm_start=False,
        max_samples=None,
    ):
        super(RandomForestQuantileRegressor, self).__init__(
            DecisionTreeQuantileRegressor(),
            n_estimators=n_estimators,
            estimator_params=(
                "criterion",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "max_features",
                "max_leaf_nodes",
                "random_state",
            ),
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            max_samples=max_samples,
        )

        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes


class ExtraTreesQuantileRegressor(BaseForestQuantileRegressor):
    """
    An extra-trees regressor that provides quantile estimates.
    This class implements a meta estimator that fits a number of
    randomized decision trees (a.k.a. extra-trees) on various sub-samples
    of the dataset and use averaging to improve the predictive accuracy
    and control over-fitting.

    Parameters
    ----------
    n_estimators : integer, optional (default=10)
        The number of trees in the forest.
    criterion : string, optional (default="squared_error")
        The function to measure the quality of a split. Supported criteria
        are "squared_error" for the mean squared error, which is equal to variance
        reduction as feature selection criterion, and "mae" for the mean
        absolute error.
        .. versionadded:: 0.18
           Mean Absolute Error (MAE) criterion.
    max_features : int, float, string or None, optional (default="auto")
        The number of features to consider when looking for the best split:
        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a percentage and
          `int(max_features * n_features)` features are considered at each
          split.
        - If "auto", then `max_features=n_features`.
        - If "sqrt", then `max_features=sqrt(n_features)`.
        - If "log2", then `max_features=log2(n_features)`.
        - If None, then `max_features=n_features`.
        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.
    max_depth : integer or None, optional (default=None)
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples.
    min_samples_split : int, float, optional (default=2)
        The minimum number of samples required to split an internal node:
        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a percentage and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.
        .. versionchanged:: 0.18
           Added float values for percentages.
    min_samples_leaf : int, float, optional (default=1)
        The minimum number of samples required to be at a leaf node:
        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a percentage and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.
        .. versionchanged:: 0.18
           Added float values for percentages.
    max_leaf_nodes : int or None, optional (default=None)
        Grow trees with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.
    bootstrap : boolean, optional (default=False)
        Whether bootstrap samples are used when building trees.
    oob_score : bool, optional (default=False)
        Whether to use out-of-bag samples to estimate the R^2 on unseen data.
    n_jobs : integer, optional (default=1)
        The number of jobs to run in parallel for both `fit` and `predict`.
        If -1, then the number of jobs is set to the number of cores.
    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.
    verbose : int, optional (default=0)
        Controls the verbosity of the tree building process.
    warm_start : bool, optional (default=False)
        When set to ``True``, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just fit a whole
        new forest.

    Attributes
    ----------
    estimators_ : list of ExtraTreeQuantileRegressor
        The collection of fitted sub-estimators.
    feature_importances_ : array of shape = [n_features]
        The feature importances (the higher, the more important the feature).
    n_features_ : int
        The number of features when ``fit`` is performed.
    n_outputs_ : int
        The number of outputs when ``fit`` is performed.
    oob_score_ : float
        Score of the training dataset obtained using an out-of-bag estimate.
    oob_prediction_ : array of shape = [n_samples]
        Prediction computed with out-of-bag estimate on the training set.
    y_train_ : array-like, shape=(n_samples,)
        Cache the target values at fit time.
    y_weights_ : array-like, shape=(n_estimators, n_samples)
        y_weights_[i, j] is the weight given to sample ``j` while
        estimator ``i`` is fit. If bootstrap is set to True, this
        reduces to a 2-D array of ones.
    y_train_leaves_ : array-like, shape=(n_estimators, n_samples)
        y_train_leaves_[i, j] provides the leaf node that y_train_[i]
        ends up when estimator j is fit. If y_train_[i] is given
        a weight of zero when estimator j is fit, then the value is -1.

    References
    ----------
    .. [1] Nicolai Meinshausen, Quantile Regression Forests
        http://www.jmlr.org/papers/volume7/meinshausen06a/meinshausen06a.pdf
    """

    def __init__(
        self,
        n_estimators=10,
        criterion="squared_error",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        max_features=1.0,
        max_leaf_nodes=None,
        bootstrap=True,
        oob_score=False,
        n_jobs=1,
        random_state=None,
        verbose=0,
        warm_start=False,
        max_samples=None,
    ):
        super(ExtraTreesQuantileRegressor, self).__init__(
            ExtraTreeQuantileRegressor(),
            n_estimators=n_estimators,
            estimator_params=(
                "criterion",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "max_features",
                "max_leaf_nodes",
                "random_state",
            ),
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            max_samples=max_samples,
        )

        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes



import copy
import logging
import os
import time
import warnings
from operator import gt, lt

from lightgbm.callback import EarlyStopException, _format_eval_result

from autogluon.common.utils.lite import disable_if_lite_mode
from autogluon.common.utils.resource_utils import ResourceManager
from autogluon.core.utils.early_stopping import SimpleES

logger = logging.getLogger(__name__)


# TODO: Add option to stop if current run's metric value is X% lower, such as min 30%, current 40% -> Stop
def early_stopping_custom(
    stopping_rounds,
    first_metric_only=False,
    metrics_to_use=None,
    start_time=None,
    time_limit=None,
    verbose=True,
    max_diff=None,
    ignore_dart_warning=False,
    manual_stop_file=None,
    train_loss_name=None,
    reporter=None,
):
    """Create a callback that activates early stopping.

    Note
    ----
    Activates early stopping.
    The model will train until the validation score stops improving.
    Validation score needs to improve at least every ``early_stopping_rounds`` round(s)
    to continue training.
    Requires at least one validation data and one metric.
    If there's more than one, will check all of them. But the training data is ignored anyway.
    To check only the first metric set ``first_metric_only`` to True.

    Parameters
    ----------
    stopping_rounds : int or tuple
       If int, The possible number of rounds without the trend occurrence.
       If tuple, contains early stopping class as first element and class init kwargs as second element.
    first_metric_only : bool, optional (default=False)
       Whether to use only the first metric for early stopping.
    verbose : bool, optional (default=True)
        Whether to print message with early stopping information.
    train_loss_name : str, optional (default=None):
        Name of metric that contains training loss value.
    reporter : optional (default=None):
        reporter object from AutoGluon scheduler.

    Returns
    -------
    callback : function
        The callback that activates early stopping.
    """
    best_score = []
    best_iter = []
    best_score_list = []
    best_trainloss = []  # stores training losses at corresponding best_iter
    cmp_op = []
    enabled = [True]
    indices_to_check = []
    init_mem_rss = []
    init_mem_avail = []
    es = []

    mem_status = ResourceManager.get_process()

    def _init(env):
        if not ignore_dart_warning:
            enabled[0] = not any((boost_alias in env.params and env.params[boost_alias] == "dart") for boost_alias in ("boosting", "boosting_type", "boost"))
        if not enabled[0]:
            warnings.warn("Early stopping is not available in dart mode")
            return
        if not env.evaluation_result_list:
            raise ValueError("For early stopping, " "at least one dataset and eval metric is required for evaluation")

        if verbose:
            msg = "Training until validation scores don't improve for {} rounds."
            logger.debug(msg.format(stopping_rounds))
            if manual_stop_file:
                logger.debug("Manually stop training by creating file at location: ", manual_stop_file)

        if isinstance(stopping_rounds, int):
            es_template = SimpleES(patience=stopping_rounds)
        else:
            es_template = stopping_rounds[0](**stopping_rounds[1])

        for eval_ret in env.evaluation_result_list:
            best_iter.append(0)
            best_score_list.append(None)
            best_trainloss.append(None)
            es.append(copy.deepcopy(es_template))
            if eval_ret[3]:
                best_score.append(float("-inf"))
                cmp_op.append(gt)
            else:
                best_score.append(float("inf"))
                cmp_op.append(lt)

        if metrics_to_use is None:
            for i in range(len(env.evaluation_result_list)):
                indices_to_check.append(i)
                if first_metric_only:
                    break
        else:
            for i, eval in enumerate(env.evaluation_result_list):
                if (eval[0], eval[1]) in metrics_to_use:
                    indices_to_check.append(i)
                    if first_metric_only:
                        break

        @disable_if_lite_mode()
        def _init_mem():
            init_mem_rss.append(mem_status.memory_info().rss)
            init_mem_avail.append(ResourceManager.get_available_virtual_mem())

        _init_mem()

    @disable_if_lite_mode()
    def _mem_early_stop():
        available = ResourceManager.get_available_virtual_mem()
        cur_rss = mem_status.memory_info().rss

        if cur_rss < init_mem_rss[0]:
            init_mem_rss[0] = cur_rss
        estimated_model_size_mb = (cur_rss - init_mem_rss[0]) >> 20
        available_mb = available >> 20

        if available_mb != 0:
            model_size_memory_ratio = estimated_model_size_mb / available_mb
        else:
            model_size_memory_ratio = 100

        if verbose or (model_size_memory_ratio > 0.25):
            logger.debug("Available Memory: " + str(available_mb) + " MB")
            logger.debug("Estimated Model Size: " + str(estimated_model_size_mb) + " MB")

        early_stop = False
        # FIXME: during parallel fits, model only knows its own memory usage and the overall system memory.
        #  Because memory usage can spike during saving, OOM can occur if many models finish at the same time and spike in memory at the same time during save.
        #  To fix this, we need to provide the per-model memory limit as a constraint passed to this method,
        #  so that we can ensure a given model isn't exceeding its portion of the memory budget.
        #  Ditto for XGBoost and CatBoost (ex: "kropt" dataset with 8-fold bagging and 32 GB memory. Fits 10k iterations on all 8 folds, then goes OOM)
        #  We also need to estimate the peak memory usage given the estimated_model_size_mb if we were to save. Otherwise we will go OOM during save anyways.
        if model_size_memory_ratio > 0.66:
            logger.warning("Warning: Large GBM model size may cause OOM error if training continues")
            logger.warning("Available Memory: " + str(available_mb) + " MB")
            logger.warning("Estimated GBM model size: " + str(estimated_model_size_mb) + " MB")
            early_stop = True

        # TODO: We will want to track size of model as well, even if we early stop before OOM, we will still crash when saving if the model is large enough
        if available_mb < 512:  # Less than 500 MB
            logger.warning("Warning: Low available memory may cause OOM error if training continues")
            logger.warning("Available Memory: " + str(available_mb) + " MB")
            logger.warning("Estimated GBM model size: " + str(estimated_model_size_mb) + " MB")
            early_stop = True

        if early_stop:
            logger.warning(
                "Warning: Early stopped GBM model prior to optimal result to avoid OOM error. Please increase available memory to avoid subpar model quality."
            )
            logger.log(
                15,
                "Early stopping, best iteration is:\n[%d]\t%s"
                % (best_iter[0] + 1, "\t".join([_format_eval_result(x, show_stdv=False) for x in best_score_list[0]])),
            )
            raise EarlyStopException(best_iter[0], best_score_list[0])

    def _callback(env):
        if not cmp_op:
            _init(env)
        if not enabled[0]:
            return
        if train_loss_name is not None:
            train_loss_evals = [eval for eval in env.evaluation_result_list if eval[0] == "train_set" and eval[1] == train_loss_name]
            train_loss_val = train_loss_evals[0][2]
        else:
            train_loss_val = 0.0
        for i in indices_to_check:
            is_best_iter = False
            eval_result = env.evaluation_result_list[i]
            _, eval_metric, score, greater_is_better = eval_result
            if best_score_list[i] is None or cmp_op[i](score, best_score[i]):
                is_best_iter = True
                best_score[i] = score
                best_iter[i] = env.iteration
                best_score_list[i] = env.evaluation_result_list
                best_trainloss[i] = train_loss_val
            if reporter is not None:  # Report current best scores for iteration, used in HPO
                if i == indices_to_check[0]:  # TODO: documentation needs to note that we assume 0th index is the 'official' validation performance metric.
                    if cmp_op[i] == gt:
                        validation_perf = score
                    else:
                        validation_perf = -score
                    reporter(
                        epoch=env.iteration + 1,
                        validation_performance=validation_perf,
                        train_loss=best_trainloss[i],
                        best_iter_sofar=best_iter[i] + 1,
                        best_valperf_sofar=best_score[i],
                        eval_metric=eval_metric,  # eval_metric here is the stopping_metric from LGBModel
                        greater_is_better=greater_is_better,
                    )
            early_stop = es[i].update(cur_round=env.iteration, is_best=is_best_iter)
            if early_stop:
                if verbose:
                    logger.log(
                        15,
                        "Early stopping, best iteration is:\n[%d]\t%s"
                        % (best_iter[i] + 1, "\t".join([_format_eval_result(x, show_stdv=False) for x in best_score_list[i]])),
                    )
                raise EarlyStopException(best_iter[i], best_score_list[i])
            elif (max_diff is not None) and (abs(score - best_score[i]) > max_diff):
                if verbose:
                    logger.debug("max_diff breached!")
                    logger.debug(abs(score - best_score[i]))
                    logger.log(
                        15,
                        "Early stopping, best iteration is:\n[%d]\t%s"
                        % (best_iter[i] + 1, "\t".join([_format_eval_result(x, show_stdv=False) for x in best_score_list[i]])),
                    )
                raise EarlyStopException(best_iter[i], best_score_list[i])
            if env.iteration == env.end_iteration - 1:
                if verbose:
                    logger.log(
                        15,
                        "Did not meet early stopping criterion. Best iteration is:\n[%d]\t%s"
                        % (best_iter[i] + 1, "\t".join([_format_eval_result(x, show_stdv=False) for x in best_score_list[i]])),
                    )
                raise EarlyStopException(best_iter[i], best_score_list[i])
            if verbose:
                logger.debug((env.iteration - best_iter[i], eval_result))
        if manual_stop_file:
            if os.path.exists(manual_stop_file):
                i = indices_to_check[0]
                logger.log(
                    20,
                    "Found manual stop file, early stopping. Best iteration is:\n[%d]\t%s"
                    % (best_iter[i] + 1, "\t".join([_format_eval_result(x, show_stdv=False) for x in best_score_list[i]])),
                )
                raise EarlyStopException(best_iter[i], best_score_list[i])
        if time_limit:
            time_elapsed = time.time() - start_time
            time_left = time_limit - time_elapsed
            if time_left <= 0:
                i = indices_to_check[0]
                logger.log(
                    20,
                    "\tRan out of time, early stopping on iteration "
                    + str(env.iteration + 1)
                    + ". Best iteration is:\n\t[%d]\t%s" % (best_iter[i] + 1, "\t".join([_format_eval_result(x, show_stdv=False) for x in best_score_list[i]])),
                )
                raise EarlyStopException(best_iter[i], best_score_list[i])

        # TODO: Add toggle parameter to early_stopping to disable this
        # TODO: Identify optimal threshold values for early_stopping based on lack of memory
        if env.iteration % 10 == 0:
            _mem_early_stop()

    _callback.order = 30
    return _callback



from __future__ import annotations

import numpy as np
from sklearn.model_selection import StratifiedKFold, train_test_split

from ..._internal.config.enums import Task

def make_dataset_split(x: np.ndarray, y: np.ndarray, task: Task, seed: int) -> tuple[np.ndarray, ...]:
    # Splits the dataset into train and validation sets with ratio 80/20

    if task == Task.REGRESSION:
        return make_standard_dataset_split(x, y, seed=seed)

    size_of_smallest_class = np.min(np.bincount(y))

    if size_of_smallest_class >= 5:
        # stratification needs have at least 5 samples in each class if split is 80/20
        return make_stratified_dataset_split(x, y, seed=seed)
    else:
        return make_standard_dataset_split(x, y, seed=seed)
    

def make_stratified_dataset_split(x, y, n_splits=5, seed=0):
    if isinstance(seed, int):
        seed = np.random.RandomState(seed)

    # Stratify doesn't shuffle the data, so we shuffle it first
    permutation = seed.permutation(len(y))
    x, y = x[permutation], y[permutation]

    min_samples_per_class = np.min(np.bincount(y))
    
    # Adjust n_splits based on both total samples and minimum samples per class
    n_samples = len(y)
    max_possible_splits = min(n_samples - 1, min_samples_per_class)
    n_splits = min(n_splits, max_possible_splits)
    
    # Ensure we have at least 2 splits if possible
    if n_samples >= 2 and min_samples_per_class >= 2:
        n_splits = max(2, n_splits)
    else:
        # If we can't do stratified splitting, fall back to standard split
        return make_standard_dataset_split(x, y, seed)

    skf = StratifiedKFold(n_splits=n_splits, shuffle=True, random_state=seed)
    indices = next(skf.split(x, y))
    x_t_train, x_t_valid = x[indices[0]], x[indices[1]] # 80%, 20%
    y_t_train, y_t_valid = y[indices[0]], y[indices[1]]

    return x_t_train, x_t_valid, y_t_train, y_t_valid


def make_standard_dataset_split(x, y, seed):
        
    return train_test_split(
        x, y, test_size=0.2, random_state=seed,
    )


import logging

from autogluon.common.utils.try_import import try_import_rapids_cuml
from autogluon.core.constants import REGRESSION

from .._utils.rapids_utils import RapidsModelMixin
from .hyperparameters.parameters import get_param_baseline
from .lr_model import LinearModel

logger = logging.getLogger(__name__)


# FIXME: If rapids is installed, normal CPU LinearModel crashes.
class LinearRapidsModel(RapidsModelMixin, LinearModel):
    """
    RAPIDS Linear model : https://rapids.ai/start.html

    NOTE: This code is experimental, it is recommend to not use this unless you are a developer.
    This was tested on rapids-21.06 via:

    conda create -n rapids-21.06 -c rapidsai -c nvidia -c conda-forge rapids=21.06 python=3.8 cudatoolkit=11.2
    conda activate rapids-21.06
    pip install --pre autogluon.tabular[all]
    """

    def _get_model_type(self):
        penalty = self.params.get("penalty", "L2")
        try_import_rapids_cuml()
        from cuml.linear_model import Lasso, LogisticRegression, Ridge

        if self.problem_type == REGRESSION:
            if penalty == "L2":
                model_type = Ridge
            elif penalty == "L1":
                model_type = Lasso
            else:
                raise AssertionError(f'Unknown value for penalty "{penalty}" - supported types are ["L1", "L2"]')
        else:
            model_type = LogisticRegression
        return model_type

    def _set_default_params(self):
        default_params = {"fit_intercept": True, "max_iter": 10000}
        if self.problem_type != REGRESSION:
            default_params.update({"solver": "qn"})
        default_params.update(get_param_baseline())
        for param, val in default_params.items():
            self._set_default_param_value(param, val)

    def _preprocess(self, X, **kwargs):
        X = super()._preprocess(X=X, **kwargs)
        if hasattr(X, 'toarray'):  # Check if it's a sparse matrix
            X = X.toarray()
        return X

    def _fit(self, X, y, **kwargs):
        """
        Custom fit method for RAPIDS cuML models that handles parameter compatibility
        and bypasses sklearn-specific incremental training approach.
        """
        # Preprocess data
        X = self.preprocess(X, is_train=True)
        if self.problem_type == 'binary':
            y = y.astype(int).values

        # Create cuML model with filtered parameters
        model_cls = self._get_model_type()

        # Comprehensive parameter filtering for cuML compatibility
        cuml_incompatible_params = {
            # AutoGluon-specific preprocessing parameters
            'vectorizer_dict_size', 'proc.ngram_range', 'proc.skew_threshold',
            'proc.impute_strategy', 'handle_text',
            # sklearn-specific parameters not supported by cuML
            'n_jobs', 'warm_start', 'multi_class', 'dual', 'intercept_scaling',
            'class_weight', 'random_state', 'verbose',
            # Parameters that need conversion or special handling
            'penalty', 'C'
        }

        # Filter out incompatible parameters
        filtered_params = {k: v for k, v in self.params.items()
                          if k not in cuml_incompatible_params}

        # Handle parameter conversions for cuML
        if self.problem_type == REGRESSION:
            # Convert sklearn's C parameter to cuML's alpha
            if 'C' in self.params:
                filtered_params['alpha'] = 1.0 / self.params['C']
        else:
            # For classification, keep C parameter
            if 'C' in self.params:
                filtered_params['C'] = self.params['C']

        # Create and fit cuML model - let cuML handle its own error messages
        self.model = model_cls(**filtered_params)
        self.model.fit(X, y)

        # Add missing sklearn-compatible attributes for AutoGluon compatibility
        self.model.n_iter_ = None  # cuML doesn't track iterations like sklearn



import random
import numpy as np
import torch


def set_seed(seed: int) -> None:
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

def seed_worker(worker_id: int) -> None:
    worker_seed = torch.initial_seed() % 2**32
    set_seed(worker_seed)


from __future__ import annotations

import copy
import os
import pandas as pd
import shutil
import uuid
from typing import Any, Type

from autogluon.common.utils.path_converter import PathConverter
from autogluon.core.constants import BINARY, MULTICLASS, REGRESSION
from autogluon.core.metrics import METRICS
from autogluon.core.models import AbstractModel, BaggedEnsembleModel
from autogluon.core.stacked_overfitting.utils import check_stacked_overfitting_from_leaderboard
from autogluon.core.utils import download, generate_train_test_split_combined, infer_problem_type, unzip

from autogluon.tabular import TabularDataset, TabularPredictor
from autogluon.tabular.testing.generate_datasets import (
    generate_toy_binary_dataset,
    generate_toy_binary_10_dataset,
    generate_toy_multiclass_dataset,
    generate_toy_regression_dataset,
    generate_toy_quantile_dataset,
    generate_toy_quantile_single_level_dataset,
    generate_toy_multiclass_10_dataset,
    generate_toy_regression_10_dataset,
    generate_toy_quantile_10_dataset,
    generate_toy_multiclass_30_dataset,
)


class DatasetLoaderHelper:
    dataset_info_dict = dict(
        # Binary dataset
        adult={
            "url": "https://autogluon.s3.amazonaws.com/datasets/AdultIncomeBinaryClassification.zip",
            "name": "AdultIncomeBinaryClassification",
            "problem_type": BINARY,
            "label": "class",
        },
        # Multiclass big dataset with 7 classes, all features are numeric. Runs SLOW.
        covertype={
            "url": "https://autogluon.s3.amazonaws.com/datasets/CoverTypeMulticlassClassification.zip",
            "name": "CoverTypeMulticlassClassification",
            "problem_type": MULTICLASS,
            "label": "Cover_Type",
        },
        # Subset of covertype dataset with 3k train/test rows. Ratio of labels is preserved.
        covertype_small={
            "url": "https://autogluon.s3.amazonaws.com/datasets/CoverTypeMulticlassClassificationSmall.zip",
            "name": "CoverTypeMulticlassClassificationSmall",
            "problem_type": MULTICLASS,
            "label": "Cover_Type",
        },
        # Regression with mixed feature-types, skewed Y-values.
        ames={
            "url": "https://autogluon.s3.amazonaws.com/datasets/AmesHousingPriceRegression.zip",
            "name": "AmesHousingPriceRegression",
            "problem_type": REGRESSION,
            "label": "SalePrice",
        },
        # Regression with multiple text field and categorical
        sts={
            "url": "https://autogluon-text.s3.amazonaws.com/glue_sts.zip",
            "name": "glue_sts",
            "problem_type": REGRESSION,
            "label": "score",
        },
    )

    toy_map = dict(
        toy_binary=generate_toy_binary_dataset,
        toy_multiclass=generate_toy_multiclass_dataset,
        toy_regression=generate_toy_regression_dataset,
        toy_quantile=generate_toy_quantile_dataset,
        toy_quantile_single_level=generate_toy_quantile_single_level_dataset,
        toy_binary_10=generate_toy_binary_10_dataset,
        toy_multiclass_10=generate_toy_multiclass_10_dataset,
        toy_regression_10=generate_toy_regression_10_dataset,
        toy_quantile_10=generate_toy_quantile_10_dataset,
        toy_multiclass_30=generate_toy_multiclass_30_dataset,
    )

    @staticmethod
    def load_dataset(name: str, directory_prefix: str = "./datasets/") -> tuple[pd.DataFrame, pd.DataFrame, dict]:
        if name in DatasetLoaderHelper.toy_map:
            return DatasetLoaderHelper.toy_map[name]()
        dataset_info = copy.deepcopy(DatasetLoaderHelper.dataset_info_dict[name])
        train_file = dataset_info.pop("train_file", "train_data.csv")
        test_file = dataset_info.pop("test_file", "test_data.csv")
        name_inner = dataset_info.pop("name")
        url = dataset_info.pop("url", None)
        train_data, test_data = DatasetLoaderHelper.load_data(
            directory_prefix=directory_prefix,
            train_file=train_file,
            test_file=test_file,
            name=name_inner,
            url=url,
        )

        return train_data, test_data, dataset_info

    # TODO: Refactor this eventually, this is old code from 2019 that can be improved (use consistent local path for datasets, don't assume zip files, etc.)
    @staticmethod
    def load_data(
        directory_prefix: str,
        train_file: str,
        test_file: str,
        name: str,
        url: str | None = None,
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Will check if files exist locally for:
            directory_prefix/name/train_file
            directory_prefix/name/test_file
        If either don't exist, then download the files from the `url` location.
        Then, load both train and test data and return them.

        Parameters
        ----------
        directory_prefix
        train_file
        test_file
        name
        url

        Returns
        -------
        train_data: pd.DataFrame
        test_data: pd.DataFrame
        """
        if not os.path.exists(directory_prefix):
            os.mkdir(directory_prefix)
        directory = directory_prefix + name + "/"
        train_file_path = directory + train_file
        test_file_path = directory + test_file
        if (not os.path.exists(train_file_path)) or (not os.path.exists(test_file_path)):
            # fetch files from s3:
            print("%s data not found locally, so fetching from %s" % (name, url))
            zip_name = download(url, directory_prefix)
            unzip(zip_name, directory_prefix)
            os.remove(zip_name)

        train_data = TabularDataset(train_file_path)
        test_data = TabularDataset(test_file_path)
        return train_data, test_data


class FitHelper:
    """
    Helper functions to test and verify predictors and models when fit through TabularPredictor's API.
    """
    @staticmethod
    def fit_and_validate_dataset(
        dataset_name: str,
        fit_args: dict[str, Any],
        init_args: dict[str, Any] | None = None,
        sample_size: int | None = 1000,  # FIXME: default to None
        refit_full: bool = True,
        delete_directory: bool = True,
        extra_metrics: list[str] | None = None,
        extra_info: bool = False,
        predictor_info: bool = False,
        expected_model_count: int | None = 2,
        fit_weighted_ensemble: bool = True,
        min_cls_count_train: int = 1,
        path_as_absolute: bool = False,
        compile: bool = False,
        compiler_configs: dict | None = None,
        allowed_dataset_features: list[str] | None = None,
        expected_stacked_overfitting_at_test: bool | None = None,
        expected_stacked_overfitting_at_val: bool | None = None,
        scikit_api: bool = False,
        use_test_data: bool = False,
        use_test_for_val: bool = False,
        raise_on_model_failure: bool | None = None,
        deepcopy_fit_args: bool = True,
    ) -> TabularPredictor:
        if compiler_configs is None:
            compiler_configs = {}
        directory_prefix = "./datasets/"
        train_data, test_data, dataset_info = DatasetLoaderHelper.load_dataset(name=dataset_name, directory_prefix=directory_prefix)
        label = dataset_info["label"]
        problem_type = dataset_info["problem_type"]
        _init_args = dict(
            label=label,
            problem_type=problem_type,
        )
        if "init_kwargs" in dataset_info:
            _init_args.update(dataset_info["init_kwargs"])
        if allowed_dataset_features is not None:
            train_data = train_data[allowed_dataset_features + [label]]
            test_data = test_data[allowed_dataset_features + [label]]

        if init_args is None:
            init_args = _init_args
        else:
            init_args = copy.deepcopy(init_args)
            _init_args.update(init_args)
            init_args = _init_args
        if "path" not in init_args:
            init_args["path"] = os.path.join(directory_prefix, dataset_name, f"AutogluonOutput_{uuid.uuid4()}")
        if path_as_absolute:
            init_args["path"] = PathConverter.to_absolute(path=init_args["path"])
            assert PathConverter._is_absolute(path=init_args["path"])
        save_path = init_args["path"]

        if deepcopy_fit_args:
            fit_args = copy.deepcopy(fit_args)
        if use_test_data:
            fit_args["test_data"] = test_data
            if use_test_for_val:
                fit_args["tuning_data"] = test_data
        if raise_on_model_failure is not None and "raise_on_model_failure" not in fit_args:
            fit_args["raise_on_model_failure"] = raise_on_model_failure
        if "fit_weighted_ensemble" not in fit_args:
            if not fit_weighted_ensemble and expected_model_count is not None:
                expected_model_count -= 1
            fit_args["fit_weighted_ensemble"] = fit_weighted_ensemble

        predictor: TabularPredictor = FitHelper.fit_dataset(
            train_data=train_data,
            init_args=init_args,
            fit_args=fit_args,
            sample_size=sample_size,
            scikit_api=scikit_api,
            min_cls_count_train=min_cls_count_train,
        )
        if compile:
            predictor.compile(models="all", compiler_configs=compiler_configs)
            predictor.persist(models="all")
        if sample_size is not None and sample_size < len(test_data):
            test_data = test_data.sample(n=sample_size, random_state=0)
        predictor.predict(test_data)
        predictor.evaluate(test_data)

        if predictor.can_predict_proba:
            pred_proba = predictor.predict_proba(test_data)
            predictor.evaluate_predictions(y_true=test_data[label], y_pred=pred_proba)
        else:
            try:
                predictor.predict_proba(test_data)
            except AssertionError:
                pass  # expected
            else:
                raise AssertionError("Expected `predict_proba` to raise AssertionError, but it didn't!")

        model_names = predictor.model_names()
        model_name = model_names[0]
        if expected_model_count is not None:
            assert len(model_names) == expected_model_count
        if refit_full:
            refit_model_names = predictor.refit_full()
            if expected_model_count is not None:
                assert len(refit_model_names) == expected_model_count
            refit_model_name = refit_model_names[model_name]
            assert "_FULL" in refit_model_name
            predictor.predict(test_data, model=refit_model_name)
            if predictor.can_predict_proba:
                predictor.predict_proba(test_data, model=refit_model_name)

            # verify that val_in_fit is False if the model supports refit_full
            model = predictor._trainer.load_model(refit_model_name)
            if isinstance(model, BaggedEnsembleModel):
                model = model.load_child(model.models[0])
            model_info = model.get_info()
            can_refit_full = model._get_tags()["can_refit_full"]
            if can_refit_full:
                assert not model_info["val_in_fit"], f"val data must not be present in refit model if `can_refit_full=True`. Maybe an exception occurred?"
            else:
                assert model_info["val_in_fit"], f"val data must be present in refit model if `can_refit_full=False`"

        if predictor_info:
            predictor.info()
        lb_kwargs = {}
        if extra_info:
            lb_kwargs["extra_info"] = True
        lb = predictor.leaderboard(test_data, extra_metrics=extra_metrics, **lb_kwargs)
        stacked_overfitting_assert(lb, predictor, expected_stacked_overfitting_at_val, expected_stacked_overfitting_at_test)

        predictor_load = predictor.load(path=predictor.path)
        predictor_load.predict(test_data)

        assert os.path.realpath(save_path) == os.path.realpath(predictor.path)
        if delete_directory:
            shutil.rmtree(save_path, ignore_errors=True)  # Delete AutoGluon output directory to ensure runs' information has been removed.
        return predictor

    @staticmethod
    def load_dataset(name: str, directory_prefix: str = "./datasets/") -> tuple[pd.DataFrame, pd.DataFrame, dict]:
        return DatasetLoaderHelper.load_dataset(name=name, directory_prefix=directory_prefix)

    @staticmethod
    def fit_dataset(
        train_data: pd.DataFrame,
        init_args: dict[str, Any],
        fit_args: dict[str, Any],
        sample_size: int | None = None,
        min_cls_count_train: int = 1,
        scikit_api: bool = False,
    ) -> TabularPredictor:
        if "problem_type" in init_args:
            problem_type = init_args["problem_type"]
        else:
            problem_type = infer_problem_type(train_data[init_args["label"]])

        if sample_size is not None and sample_size < len(train_data):
            train_data, _ = generate_train_test_split_combined(
                data=train_data,
                label=init_args["label"],
                problem_type=problem_type,
                test_size=len(train_data) - sample_size,
                min_cls_count_train=min_cls_count_train,
            )

        if scikit_api:
            from autogluon.tabular.experimental import TabularClassifier, TabularRegressor

            X = train_data.drop(columns=[init_args["label"]])
            y = train_data[init_args["label"]]
            if problem_type in [REGRESSION]:
                regressor = TabularRegressor(init_args=init_args, fit_args=fit_args)
                regressor.fit(X, y)
                return regressor.predictor_
            else:
                classifier = TabularClassifier(init_args=init_args, fit_args=fit_args)
                classifier.fit(X, y)
                return classifier.predictor_
        else:
            return TabularPredictor(**init_args).fit(train_data, **fit_args)

    @staticmethod
    def verify_model(
        model_cls: Type[AbstractModel],
        model_hyperparameters: dict[str, Any],
        bag: bool | str = "first",
        refit_full: bool | str = "first",
        extra_metrics: bool = False,
        require_known_problem_types: bool = True,
        raise_on_model_failure: bool = True,
        problem_types: list[str] | None = None,
        **kwargs,
    ):
        """

        Parameters
        ----------
        model_cls
        model_hyperparameters
        bag
        refit_full
        extra_metrics
        require_known_problem_types
        raise_on_model_failure
        problem_types: list[str], optional
            If specified, checks the given problem_types.
            If None, checks `model_cls.supported_problem_types()`
        **kwargs

        Returns
        -------

        """
        fit_args = dict(
            hyperparameters={model_cls: model_hyperparameters},
        )
        supported_problem_types = model_cls.supported_problem_types()
        if supported_problem_types is None:
            raise AssertionError(
                f"Model must specify `cls.supported_problem_types`"
                f"""\nExample code:
    @classmethod
    def supported_problem_types(cls) -> list[str] | None:
        return ["binary", "multiclass", "regression", "quantile"]
        """
            )
        assert isinstance(supported_problem_types, list)
        assert len(supported_problem_types) > 0

        known_problem_types = [
            "binary",
            "multiclass",
            "regression",
            "quantile",
            "softclass",
        ]

        if require_known_problem_types:
            for problem_type in supported_problem_types:
                if problem_type not in known_problem_types:
                    raise AssertionError(
                        f"Model {model_cls.__name__} supports an unknown problem_type: {problem_type}"
                        f"\nKnown problem types: {known_problem_types}"
                        f"\nEither remove the unknown problem_type from `model_cls.supported_problem_types` or set `require_known_problem_types=False`"
                    )

        problem_type_dataset_map = {
            "binary": ["toy_binary"],
            "multiclass": ["toy_multiclass"],
            "regression": ["toy_regression"],
            "quantile": ["toy_quantile", "toy_quantile_single_level"],
        }

        problem_types_refit_full = []
        if refit_full:
            if isinstance(refit_full, bool):
                problem_types_refit_full = supported_problem_types
            elif refit_full == "first":
                problem_types_refit_full = supported_problem_types[:1]

        if problem_types is None:
            problem_types_to_check = supported_problem_types
        else:
            problem_types_to_check = problem_types

        for problem_type in problem_types_to_check:
            if problem_type not in problem_type_dataset_map:
                print(f"WARNING: Skipping check on problem_type='{problem_type}': No dataset available")
                continue
            _extra_metrics = None
            if extra_metrics:
                _extra_metrics = METRICS.get(problem_type, None)
            refit_full = problem_type in problem_types_refit_full
            for dataset_name in problem_type_dataset_map[problem_type]:
                FitHelper.fit_and_validate_dataset(
                    dataset_name=dataset_name,
                    fit_args=fit_args,
                    fit_weighted_ensemble=False,
                    refit_full=refit_full,
                    extra_metrics=_extra_metrics,
                    raise_on_model_failure=raise_on_model_failure,
                    **kwargs,
                )

        if bag:
            model_params_bag = copy.deepcopy(model_hyperparameters)
            model_params_bag["ag.ens.fold_fitting_strategy"] = "sequential_local"
            fit_args_bag = dict(
                hyperparameters={model_cls: model_params_bag},
                num_bag_folds=2,
                num_bag_sets=1,
            )
            if isinstance(bag, bool):
                problem_types_bag = problem_types_to_check
            elif bag == "first":
                problem_types_bag = problem_types_to_check[:1]
            else:
                raise ValueError(f"Unknown 'bag' value: {bag}")

            for problem_type in problem_types_bag:
                _extra_metrics = None
                if extra_metrics:
                    _extra_metrics = METRICS.get(problem_type, None)
                refit_full = problem_type in problem_types_refit_full
                for dataset_name in problem_type_dataset_map[problem_type]:
                    FitHelper.fit_and_validate_dataset(
                        dataset_name=dataset_name,
                        fit_args=fit_args_bag,
                        fit_weighted_ensemble=False,
                        refit_full=refit_full,
                        extra_metrics=_extra_metrics,
                        raise_on_model_failure=raise_on_model_failure,
                        **kwargs,
                    )


def stacked_overfitting_assert(
    lb: pd.DataFrame,
    predictor: TabularPredictor,
    expected_stacked_overfitting_at_val: bool | None,
    expected_stacked_overfitting_at_test: bool | None,
):
    if expected_stacked_overfitting_at_val is not None:
        assert predictor._stacked_overfitting_occurred == expected_stacked_overfitting_at_val, "Expected stacked overfitting at val mismatch!"

    if expected_stacked_overfitting_at_test is not None:
        stacked_overfitting = check_stacked_overfitting_from_leaderboard(lb)
        assert stacked_overfitting == expected_stacked_overfitting_at_test, "Expected stacked overfitting at test mismatch!"



from __future__ import annotations

import time
import warnings
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from autogluon.core.constants import BINARY, MULTICLASS, REGRESSION
from autogluon.core.models import AbstractModel

from .hyperparameters.parameters import get_param_baseline
from .hyperparameters.searchspaces import get_default_searchspace

if TYPE_CHECKING:
    from autogluon.core.metrics import Scorer


class EbmCallback:
    """Time limit callback for EBM."""

    def __init__(self, seconds: float):
        self.seconds = seconds
        self.end_time: float | None = None

    def __call__(self, *args, **kwargs):
        if self.end_time is None:
            self.end_time = time.monotonic() + self.seconds
            return False
        return time.monotonic() > self.end_time


class EBMModel(AbstractModel):
    """
    The Explainable Boosting Machine (EBM) is a glass-box generalized additive model
    with automatic interaction detection (https://interpret.ml/docs). EBMs are
    designed to be highly interpretable while achieving accuracy comparable to
    black-box models on a wide range of tabular datasets.

    Requires the 'interpret' or 'interpret-core' package. Install via:
    
    pip install interpret


    Paper: InterpretML: A Unified Framework for Machine Learning Interpretability
    
    Authors: H. Nori, S. Jenkins, P. Koch, and R. Caruana 2019
    
    Codebase: https://github.com/interpretml/interpret

    License: MIT

    .. versionadded:: 1.5.0
    """

    ag_key = "EBM"
    ag_name = "EBM"
    ag_priority = 35
    
    def _fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        X_val: pd.DataFrame | None = None,
        y_val: pd.Series | None = None,
        time_limit: float | None = None,
        sample_weight: np.ndarray | None = None,
        sample_weight_val: np.ndarray | None = None,
        num_cpus: int | str = "auto",
        **kwargs,
    ):
        # Preprocess data.
        X = self.preprocess(X)
        if X_val is not None:
            X_val = self.preprocess(X_val)

        features = self._features
        if features is None:
            features = X.columns

        params = construct_ebm_params(
            self.problem_type,
            self._get_model_params(),
            features,
            self.stopping_metric,
            num_cpus,
            time_limit,
        )

        # Init Class
        model_cls = get_class_from_problem_type(self.problem_type)
        self.model = model_cls(random_state=self.random_seed, **params)

        # Handle validation data format for EBM
        fit_X = X
        fit_y = y
        fit_sample_weight = sample_weight
        bags = None
        if X_val is not None:
            fit_X = pd.concat([X, X_val], ignore_index=True)
            fit_y = pd.concat([y, y_val], ignore_index=True)
            if sample_weight is not None:
                fit_sample_weight = np.hstack([sample_weight, sample_weight_val])
            bags = np.full((len(fit_X), 1), 1, np.int8)
            bags[len(X) :, 0] = -1

        with warnings.catch_warnings():  # try to filter joblib warnings
            warnings.filterwarnings(
                "ignore",
                category=UserWarning,
                message=".*resource_tracker: process died.*",
            )
            self.model.fit(fit_X, fit_y, sample_weight=fit_sample_weight, bags=bags)

    def _get_random_seed_from_hyperparameters(
        self, hyperparameters: dict
    ) -> int | None | str:
        return hyperparameters.get("random_state", "N/A")

    def _set_default_params(self):
        default_params = get_param_baseline(problem_type=self.problem_type, num_classes=self.num_classes)
        for param, val in default_params.items():
            self._set_default_param_value(param, val)

    def _get_default_searchspace(self):
        return get_default_searchspace(problem_type=self.problem_type, num_classes=self.num_classes)

    def _get_default_auxiliary_params(self) -> dict:
        default_auxiliary_params = super()._get_default_auxiliary_params()
        extra_auxiliary_params = {
            "valid_raw_types": ["int", "float", "category"],
        }
        default_auxiliary_params.update(extra_auxiliary_params)
        return default_auxiliary_params

    @classmethod
    def supported_problem_types(cls) -> list[str] | None:
        return ["binary", "multiclass", "regression"]

    @classmethod
    def _class_tags(cls) -> dict:
        return {"can_estimate_memory_usage_static": True}

    def _more_tags(self) -> dict:
        """EBMs support refit full."""
        return {"can_refit_full": True}

    def _estimate_memory_usage(self, X: pd.DataFrame, y: pd.Series | None = None, **kwargs) -> int:
        return self.estimate_memory_usage_static(
            X=X,
            y=y,
            hyperparameters=self._get_model_params(),
            problem_type=self.problem_type,
            num_classes=self.num_classes,
            features=self._features,
            **kwargs,
        )

    @classmethod
    def _estimate_memory_usage_static(
        cls,
        *,
        X: pd.DataFrame,
        y: pd.Series | None = None,
        hyperparameters: dict | None = None,
        problem_type: str = "infer",
        num_classes: int = 1,
        features=None,
        **kwargs,
    ) -> int:
        """Returns the expected peak memory usage in bytes of the EBM model during fit."""
        # TODO: we can improve the memory estimate slightly by using num_classes if y is None

        if features is None:
            features = X.columns

        model_cls = get_class_from_problem_type(problem_type)
        params = construct_ebm_params(problem_type, hyperparameters, features)
        baseline_memory_bytes = 400_000_000  # 400 MB baseline memory

        # assuming we call pd.concat([X, X_val], ignore_index=True), then X size will be doubled
        return baseline_memory_bytes + model_cls(**params).estimate_mem(
            X, y, data_multiplier=2.0
        )

    def _validate_fit_memory_usage(self, mem_error_threshold: float = 1, **kwargs):
        # Given the good mem estimates with overhead, we set the threshold to 1.
        return super()._validate_fit_memory_usage(
            mem_error_threshold=mem_error_threshold, **kwargs
        )


def construct_ebm_params(
    problem_type,
    hyperparameters=None,
    features=None,
    stopping_metric=None,
    num_cpus=-1,
    time_limit=None,
):
    if hyperparameters is None:
        hyperparameters = {}

    hyperparameters = hyperparameters.copy()  # we pop values below, so copy.

    # The user can specify nominal and continuous columns.
    continuous_columns = hyperparameters.pop("continuous_columns", [])
    nominal_columns = hyperparameters.pop("nominal_columns", [])

    feature_types = None
    if features is not None:
        feature_types = []
        for c in features:
            if c in continuous_columns:
                f_type = "continuous"
            elif c in nominal_columns:
                f_type = "nominal"
            else:
                f_type = "auto"
            feature_types.append(f_type)

    # Default parameters for EBM
    params = {
        "outer_bags": 1,  # AutoGluon ensemble creates outer bags, no need for this overhead.
        "n_jobs": 1,  # EBM only parallelizes across outer bags currently, so ignore num_cpus
        "feature_names": features,
        "feature_types": feature_types,
    }
    if stopping_metric is not None:
        params["objective"] = get_metric_from_ag_metric(
            metric=stopping_metric, problem_type=problem_type
        )
    if time_limit is not None:
        params["callback"] = EbmCallback(time_limit)

    params.update(hyperparameters)
    return params


def get_class_from_problem_type(problem_type: str):
    if problem_type in [BINARY, MULTICLASS]:
        from interpret.glassbox import ExplainableBoostingClassifier

        model_cls = ExplainableBoostingClassifier
    elif problem_type == REGRESSION:
        from interpret.glassbox import ExplainableBoostingRegressor

        model_cls = ExplainableBoostingRegressor
    else:
        raise ValueError(f"Unsupported problem type: {problem_type}")
    return model_cls


def get_metric_from_ag_metric(*, metric: Scorer, problem_type: str):
    """Map AutoGluon metric to EBM metric for early stopping."""
    if problem_type in [BINARY, MULTICLASS]:
        metric_class = "log_loss"
    elif problem_type == REGRESSION:
        metric_class = "rmse"
    else:
        raise AssertionError(f"EBM does not support {problem_type} problem type.")

    return metric_class



from autogluon.core.constants import BINARY, MULTICLASS, REGRESSION, SOFTCLASS

def get_param_baseline(problem_type, num_classes=None):
    if problem_type == BINARY:
        return get_param_binary_baseline()
    elif problem_type == MULTICLASS:
        return get_param_multiclass_baseline(num_classes=num_classes)
    elif problem_type == SOFTCLASS:
        return get_param_multiclass_baseline(num_classes=num_classes)
    elif problem_type == REGRESSION:
        return get_param_regression_baseline()
    else:
        return get_param_binary_baseline()


def get_base_params():
    base_params = {}
    return base_params


def get_param_binary_baseline():
    params = get_base_params()
    baseline_params = {}
    params.update(baseline_params)
    return params


def get_param_multiclass_baseline(num_classes):
    params = get_base_params()
    baseline_params = {}
    params.update(baseline_params)
    return params


def get_param_regression_baseline():
    params = get_base_params()
    baseline_params = {}
    params.update(baseline_params)
    return params



# optimized for <=10000 samples and <=500 features, with a GPU present
hyperparameter_portfolio_zeroshot_2025_small = {
    "TABPFNV2": [
        {
            "ag_args": {'name_suffix': '_r143', 'priority': -1},
            "average_before_softmax": False,
            "classification_model_path": 'tabpfn-v2-classifier-od3j1g5m.ckpt',
            "inference_config/FINGERPRINT_FEATURE": False,
            "inference_config/OUTLIER_REMOVAL_STD": None,
            "inference_config/POLYNOMIAL_FEATURES": 'no',
            "inference_config/PREPROCESS_TRANSFORMS": [{'append_original': True, 'categorical_name': 'ordinal_very_common_categories_shuffled', 'global_transformer_name': None, 'name': 'safepower', 'subsample_features': -1}, {'append_original': True, 'categorical_name': 'ordinal_very_common_categories_shuffled', 'global_transformer_name': None, 'name': 'quantile_uni', 'subsample_features': -1}],
            "inference_config/REGRESSION_Y_PREPROCESS_TRANSFORMS": [None, 'power'],
            "inference_config/SUBSAMPLE_SAMPLES": 0.99,
            "model_type": 'single',
            "n_ensemble_repeats": 4,
            "regression_model_path": 'tabpfn-v2-regressor-wyl4o83o.ckpt',
            "softmax_temperature": 0.75,
        },
        {
            "ag_args": {'name_suffix': '_r94', 'priority': -3},
            "average_before_softmax": True,
            "classification_model_path": 'tabpfn-v2-classifier-vutqq28w.ckpt',
            "inference_config/FINGERPRINT_FEATURE": True,
            "inference_config/OUTLIER_REMOVAL_STD": None,
            "inference_config/POLYNOMIAL_FEATURES": 'no',
            "inference_config/PREPROCESS_TRANSFORMS": [{'append_original': True, 'categorical_name': 'ordinal_very_common_categories_shuffled', 'global_transformer_name': None, 'name': 'quantile_uni', 'subsample_features': 0.99}],
            "inference_config/REGRESSION_Y_PREPROCESS_TRANSFORMS": [None],
            "inference_config/SUBSAMPLE_SAMPLES": None,
            "model_type": 'single',
            "n_ensemble_repeats": 4,
            "regression_model_path": 'tabpfn-v2-regressor-5wof9ojf.ckpt',
            "softmax_temperature": 0.9,
        },
        {
            "ag_args": {'name_suffix': '_r181', 'priority': -4},
            "average_before_softmax": False,
            "classification_model_path": 'tabpfn-v2-classifier-llderlii.ckpt',
            "inference_config/FINGERPRINT_FEATURE": False,
            "inference_config/OUTLIER_REMOVAL_STD": 9.0,
            "inference_config/POLYNOMIAL_FEATURES": 50,
            "inference_config/PREPROCESS_TRANSFORMS": [{'append_original': True, 'categorical_name': 'onehot', 'global_transformer_name': 'svd', 'name': 'quantile_uni_coarse', 'subsample_features': 0.99}],
            "inference_config/REGRESSION_Y_PREPROCESS_TRANSFORMS": ['power'],
            "inference_config/SUBSAMPLE_SAMPLES": None,
            "model_type": 'single',
            "n_ensemble_repeats": 4,
            "regression_model_path": 'tabpfn-v2-regressor.ckpt',
            "softmax_temperature": 0.95,
        },
    ],
    "GBM": [
        {
            "ag_args": {'name_suffix': '_r33', 'priority': -2},
            "bagging_fraction": 0.9625293420216,
            "bagging_freq": 1,
            "cat_l2": 0.1236875455555,
            "cat_smooth": 68.8584757332856,
            "extra_trees": False,
            "feature_fraction": 0.6189215809382,
            "lambda_l1": 0.1641757352921,
            "lambda_l2": 0.6937755557881,
            "learning_rate": 0.0154031028561,
            "max_cat_to_onehot": 17,
            "min_data_in_leaf": 1,
            "min_data_per_group": 30,
            "num_leaves": 68,
        },
        {
            "ag_args": {'name_suffix': '_r21', 'priority': -16},
            "bagging_fraction": 0.7218730663234,
            "bagging_freq": 1,
            "cat_l2": 0.0296205152578,
            "cat_smooth": 0.0010255271303,
            "extra_trees": False,
            "feature_fraction": 0.4557131604374,
            "lambda_l1": 0.5219704038237,
            "lambda_l2": 0.1070959487853,
            "learning_rate": 0.0055891584996,
            "max_cat_to_onehot": 71,
            "min_data_in_leaf": 50,
            "min_data_per_group": 10,
            "num_leaves": 30,
        },
        {
            "ag_args": {'name_suffix': '_r11', 'priority': -19},
            "bagging_fraction": 0.775784726514,
            "bagging_freq": 1,
            "cat_l2": 0.3888471449178,
            "cat_smooth": 0.0057144748021,
            "extra_trees": True,
            "feature_fraction": 0.7732354787904,
            "lambda_l1": 0.2211002452568,
            "lambda_l2": 1.1318405980187,
            "learning_rate": 0.0090151778542,
            "max_cat_to_onehot": 15,
            "min_data_in_leaf": 4,
            "min_data_per_group": 15,
            "num_leaves": 2,
        },
    ],
    "CAT": [
        {
            "ag_args": {'priority': -5},
        },
        {
            "ag_args": {'name_suffix': '_r51', 'priority': -10},
            "boosting_type": 'Plain',
            "bootstrap_type": 'Bernoulli',
            "colsample_bylevel": 0.8771035272558,
            "depth": 7,
            "grow_policy": 'SymmetricTree',
            "l2_leaf_reg": 2.0107286863021,
            "leaf_estimation_iterations": 2,
            "learning_rate": 0.0058424016622,
            "max_bin": 254,
            "max_ctr_complexity": 4,
            "model_size_reg": 0.1307400355809,
            "one_hot_max_size": 23,
            "subsample": 0.809527841437,
        },
        {
            "ag_args": {'name_suffix': '_r10', 'priority': -12},
            "boosting_type": 'Plain',
            "bootstrap_type": 'Bernoulli',
            "colsample_bylevel": 0.8994502668431,
            "depth": 6,
            "grow_policy": 'Depthwise',
            "l2_leaf_reg": 1.8187025215896,
            "leaf_estimation_iterations": 7,
            "learning_rate": 0.005177304142,
            "max_bin": 254,
            "max_ctr_complexity": 4,
            "model_size_reg": 0.5247386875068,
            "one_hot_max_size": 53,
            "subsample": 0.8705228845742,
        },
        {
            "ag_args": {'name_suffix': '_r24', 'priority': -15},
            "boosting_type": 'Plain',
            "bootstrap_type": 'Bernoulli',
            "colsample_bylevel": 0.8597809376276,
            "depth": 8,
            "grow_policy": 'Depthwise',
            "l2_leaf_reg": 0.3628261923976,
            "leaf_estimation_iterations": 5,
            "learning_rate": 0.016851077771,
            "max_bin": 254,
            "max_ctr_complexity": 4,
            "model_size_reg": 0.1253820547902,
            "one_hot_max_size": 20,
            "subsample": 0.8120271122061,
        },
        {
            "ag_args": {'name_suffix': '_r91', 'priority': -17},
            "boosting_type": 'Plain',
            "bootstrap_type": 'Bernoulli',
            "colsample_bylevel": 0.8959275863514,
            "depth": 4,
            "grow_policy": 'SymmetricTree',
            "l2_leaf_reg": 0.0026915894253,
            "leaf_estimation_iterations": 12,
            "learning_rate": 0.0475233791203,
            "max_bin": 254,
            "max_ctr_complexity": 5,
            "model_size_reg": 0.1633175256924,
            "one_hot_max_size": 11,
            "subsample": 0.798554178926,
        },
    ],
    "TABM": [
        {
            "ag_args": {'name_suffix': '_r184', 'priority': -6},
            "amp": False,
            "arch_type": 'tabm-mini',
            "batch_size": 'auto',
            "d_block": 864,
            "d_embedding": 24,
            "dropout": 0.0,
            "gradient_clipping_norm": 1.0,
            "lr": 0.0019256819924656217,
            "n_blocks": 3,
            "num_emb_n_bins": 3,
            "num_emb_type": 'pwl',
            "patience": 16,
            "share_training_batches": False,
            "tabm_k": 32,
            "weight_decay": 0.0,
        },
        {
            "ag_args": {'name_suffix': '_r69', 'priority': -7},
            "amp": False,
            "arch_type": 'tabm-mini',
            "batch_size": 'auto',
            "d_block": 848,
            "d_embedding": 28,
            "dropout": 0.40215621636031007,
            "gradient_clipping_norm": 1.0,
            "lr": 0.0010413640454559532,
            "n_blocks": 3,
            "num_emb_n_bins": 18,
            "num_emb_type": 'pwl',
            "patience": 16,
            "share_training_batches": False,
            "tabm_k": 32,
            "weight_decay": 0.0,
        },
        {
            "ag_args": {'name_suffix': '_r52', 'priority': -11},
            "amp": False,
            "arch_type": 'tabm-mini',
            "batch_size": 'auto',
            "d_block": 1024,
            "d_embedding": 32,
            "dropout": 0.0,
            "gradient_clipping_norm": 1.0,
            "lr": 0.0006297851297842611,
            "n_blocks": 4,
            "num_emb_n_bins": 22,
            "num_emb_type": 'pwl',
            "patience": 16,
            "share_training_batches": False,
            "tabm_k": 32,
            "weight_decay": 0.06900108498839816,
        },
        {
            "ag_args": {'priority': -13},
        },
        {
            "ag_args": {'name_suffix': '_r191', 'priority': -14},
            "amp": False,
            "arch_type": 'tabm-mini',
            "batch_size": 'auto',
            "d_block": 864,
            "d_embedding": 8,
            "dropout": 0.45321529282058803,
            "gradient_clipping_norm": 1.0,
            "lr": 0.0003781238075322413,
            "n_blocks": 4,
            "num_emb_n_bins": 27,
            "num_emb_type": 'pwl',
            "patience": 16,
            "share_training_batches": False,
            "tabm_k": 32,
            "weight_decay": 0.01766851962579851,
        },
        {
            "ag_args": {'name_suffix': '_r49', 'priority': -20},
            "amp": False,
            "arch_type": 'tabm-mini',
            "batch_size": 'auto',
            "d_block": 640,
            "d_embedding": 28,
            "dropout": 0.15296207419190627,
            "gradient_clipping_norm": 1.0,
            "lr": 0.002277678490593717,
            "n_blocks": 3,
            "num_emb_n_bins": 48,
            "num_emb_type": 'pwl',
            "patience": 16,
            "share_training_batches": False,
            "tabm_k": 32,
            "weight_decay": 0.0578159148243893,
        },
    ],
    "TABICL": [
        {
            "ag_args": {'priority': -8},
        },
    ],
    "XGB": [
        {
            "ag_args": {'name_suffix': '_r171', 'priority': -9},
            "colsample_bylevel": 0.9213705632288,
            "colsample_bynode": 0.6443385965381,
            "enable_categorical": True,
            "grow_policy": 'lossguide',
            "learning_rate": 0.0068171645251,
            "max_cat_to_onehot": 8,
            "max_depth": 6,
            "max_leaves": 10,
            "min_child_weight": 0.0507304250576,
            "reg_alpha": 4.2446346389037,
            "reg_lambda": 1.4800570021253,
            "subsample": 0.9656290596647,
        },
        {
            "ag_args": {'name_suffix': '_r40', 'priority': -18},
            "colsample_bylevel": 0.6377491713202,
            "colsample_bynode": 0.9237625621103,
            "enable_categorical": True,
            "grow_policy": 'lossguide',
            "learning_rate": 0.0112462621131,
            "max_cat_to_onehot": 33,
            "max_depth": 10,
            "max_leaves": 35,
            "min_child_weight": 0.1403464856034,
            "reg_alpha": 3.4960653958503,
            "reg_lambda": 1.3062320805235,
            "subsample": 0.6948898835178,
        },
    ],
    "MITRA": [
        {
            "n_estimators": 1,
            "fine_tune": True,
            "fine_tune_steps": 50,
            "ag.num_gpus": 1,
            "ag_args": {'priority': -21},
        },
    ],
}



import json
import logging
import os
from typing import Optional, Union

import einops
import einx
import torch
import torch.nn as nn
import torch.nn.functional as F
from huggingface_hub import hf_hub_download
from safetensors.torch import load_file, save_file

# Try to import flash attention, but make it optional
try:
    from flash_attn.bert_padding import pad_input, unpad_input
    from flash_attn.flash_attn_interface import flash_attn_varlen_func
    FLASH_ATTN_AVAILABLE = True
except ImportError:
    FLASH_ATTN_AVAILABLE = False

from torch.utils.checkpoint import checkpoint

from ..._internal.config.enums import Task
from ..._internal.models.base import BaseModel
from ..._internal.models.embedding import (
    Tab2DEmbeddingX,
    Tab2DEmbeddingYClasses,
    Tab2DEmbeddingYRegression,
    Tab2DQuantileEmbeddingX,
)

logger = logging.getLogger(__name__)


class Tab2D(BaseModel):

    def __init__(
        self,
        dim: int,
        dim_output: int,
        n_layers: int,
        n_heads: int,
        task: Union[str, Task],
        use_pretrained_weights: bool,
        path_to_weights: str,
        device: str = "cuda",  # Add device parameter
    ) -> None:

        super().__init__()

        self.dim = dim
        self.dim_output = dim_output
        self.n_layers = n_layers
        self.n_heads = n_heads
        self.task = task
        self.device_type = device

        # Determine if we can use flash attention
        self.use_flash_attn = FLASH_ATTN_AVAILABLE and device.startswith("cuda")

        if type(self.task) == str:
            self.task = Task[self.task]

        self.x_quantile = Tab2DQuantileEmbeddingX(dim)
        self.x_embedding = Tab2DEmbeddingX(dim)


        if self.task == Task.CLASSIFICATION:
            self.y_embedding = Tab2DEmbeddingYClasses(dim, dim_output)     # type: nn.Module
        elif self.task == Task.REGRESSION:
            if self.dim_output == 1:
                self.y_embedding = Tab2DEmbeddingYRegression(dim)
            else:
                self.y_embedding = Tab2DEmbeddingYClasses(dim, dim_output)
        else:
            raise ValueError(f"Task {task} not supported")

        self.layers = nn.ModuleList()

        for _ in range(n_layers):
            self.layers.append(Layer(dim, n_heads, self.use_flash_attn))

        self.final_layer_norm = nn.LayerNorm(dim)

        self.final_layer = nn.Linear(dim, dim_output, bias=True)

        if use_pretrained_weights:
            if device == "cpu":
                # For CPU, use weights_only=False since CUDA checkpoints are incompatible with weights_only=True
                self.load_state_dict(torch.load(path_to_weights, weights_only=False, map_location=torch.device('cpu')))
            else:
                # For GPU, use weights_only=True for security
                self.load_state_dict(torch.load(path_to_weights, weights_only=True, map_location=device))
        else:
            self.init_weights()


    def forward(
            self,
            x_support: torch.Tensor, # (b, n_s, f)
            y_support: torch.Tensor, # (b, n_s)
            x_query: torch.Tensor, # (b, n_q, f)
            padding_features: torch.Tensor, # (b, f), "1" represents padding, "0" represents valid values
            padding_obs_support: torch.Tensor, # (b, n_s)
            padding_obs_query__: torch.Tensor, # (b, n_q)
        ):

        """
        x_support is (batch_size, n_observations_support, n_features)
        y_support is (batch_size, n_observations_support)

        x_query is (batch_size, n_observations_query, n_features)

        returns:

        y_query is (batch_size, n_observations_query, n_classes)

        syntax:
        b = batch size
        s = number of observations
        f = number of features
        d = dimension of embedding
        c = number of classes
        """

        x_query__ = x_query

        batch_size = x_support.shape[0]
        n_obs_support = x_support.shape[1]
        n_obs_query__ = x_query__.shape[1]

        x_support, x_query__ = self.x_quantile(x_support, x_query__, padding_obs_support, padding_features)
        x_support = self.x_embedding(x_support) # (b, n_s, f, d)
        x_query__ = self.x_embedding(x_query__) # (b, n_q, f, d)
        y_support, y_query__ = self.y_embedding(y_support, padding_obs_support, n_obs_query__) # (b, n_s, 1, d), (b, n_q, 1, d)

        support, pack_support = einops.pack((y_support, x_support), 'b s * d') # (b, n_s, f+1, d)
        query__, pack_query__ = einops.pack((y_query__, x_query__), 'b s * d') # (b, n_q, f+1, d)

        padding_features_y = torch.zeros((batch_size, 1), device=padding_features.device, dtype=torch.bool) # (b, 1)
        padding_features, _ = einops.pack((padding_features_y, padding_features), 'b *') # (b, f+1)

        if self.use_flash_attn:
            padder_support = Padder(support, padding_obs_support, padding_features)
            padder_query__ = Padder(query__, padding_obs_query__, padding_features)

            support = padder_support.base_to_obs(support) # (n_valid_s, d)
            query__ = padder_query__.base_to_obs(query__)  # (n_valid_q, d)

            for layer in self.layers:
                support, query__ = checkpoint(layer, support, query__, padder_support, padder_query__, use_reentrant=False) # (n_valid_s, d), (n_valid_q, d)

            query__ = self.final_layer_norm(query__)
            query__ = self.final_layer(query__) # (n_valid_q, d)

            query__ = padder_query__.obs_to_base(query__) # (b, n_q, f+1, c)
        else:
            # For CPU/non-flash attention, work with standard tensor format
            for layer in self.layers:
                support, query__ = checkpoint(layer, support, query__, None, None,
                                            batch_size, padding_obs_support, padding_obs_query__, padding_features, use_reentrant=False)

            query__ = self.final_layer_norm(query__)
            query__ = self.final_layer(query__) # (b, n_q, f+1, c)

        y_query__, x_query__ = einops.unpack(query__, pack_query__, 'b s * c') # (b, n_q, 1, c), (b, n_q, f, c)

        if self.task == Task.REGRESSION:
            # output has shape (batch_size, n_observations_query, n_features, n_classes)
            # we want to remove the n_features dimension, and for regression, the n_classes dimension
            if self.dim_output == 1:
                y_query__ = y_query__[:, :, 0, 0]
            else:
                y_query__ = y_query__[:, :, 0, :]
        elif self.task == Task.CLASSIFICATION:
            y_query__ = y_query__[:, :, 0, :]
        else:
            raise ValueError(f"Task {self.task} not supported")

        return y_query__


    def init_weights(self) -> None:

        nn.init.normal_(self.x_embedding.x_embedding.weight, mean=0.0, std=1.0)
        nn.init.normal_(self.x_embedding.x_embedding.bias, mean=0.0, std=1.0)
        nn.init.normal_(self.y_embedding.y_embedding.weight, mean=0.0, std=1.0)
        nn.init.normal_(self.y_embedding.y_mask.weight, mean=0.0, std=1.0)

        # default PyTorch initialization for everything else


    def save_pretrained(self, save_directory: str):
        os.makedirs(save_directory, exist_ok=True)

        save_file(self.state_dict(), os.path.join(save_directory, "model.safetensors"))

        config = {
            "dim": self.dim,
            "dim_output": self.dim_output,
            "n_layers": self.n_layers,
            "n_heads": self.n_heads,
            "task": str(self.task).upper(),
        }
        with open(os.path.join(save_directory, "config.json"), "w") as f:
            json.dump(config, f)

            
    @classmethod
    def from_pretrained(cls, path_or_repo_id: str, device: str = "cuda") -> "Tab2D":

        config_path = hf_hub_download(repo_id=path_or_repo_id, filename="config.json")
        with open(config_path, "r") as f:
            config = json.load(f)

        model = cls(
            dim=config["dim"],
            dim_output=config["dim_output"],
            n_layers=config["n_layers"],
            n_heads=config["n_heads"],
            task=config["task"],
            use_pretrained_weights=False,
            path_to_weights="",
            device=device
        )

        weights_path = hf_hub_download(repo_id=path_or_repo_id, filename="model.safetensors")
        state_dict = load_file(weights_path, device=device)
        model.load_state_dict(state_dict)

        return model
    
    
class Padder(torch.nn.Module):

    def __init__(self, x: torch.Tensor, padding_mask: torch.Tensor, feature_mask: torch.Tensor) -> None:

        super().__init__()

        self.padding_mask = padding_mask
        self.feature_mask = feature_mask

        n_obs, n_feat = x.shape[1], x.shape[2]
        self.batch_size = x.shape[0]

        if not FLASH_ATTN_AVAILABLE:
            # CPU fallback: implement simplified padding logic without flash attention
            self._init_cpu_fallback(x, n_obs, n_feat)
            return

        # GPU path with flash attention
        self._init_flash_attn(x, n_obs, n_feat)

    def _init_cpu_fallback(self, x: torch.Tensor, n_obs: int, n_feat: int):
        """Initialize CPU-compatible version without flash attention dependencies."""
        # For CPU, we don't need the complex unpadding/padding logic
        # We'll implement pass-through methods that preserve tensor shapes
        self.cpu_mode = True

        # Store original shapes for reference
        self.original_shape = x.shape
        self.n_obs = n_obs
        self.n_feat = n_feat

        # These attributes won't be used in CPU mode but need to exist for compatibility
        self.cu_seqlens_o = None
        self.cu_seqlens_f = None
        self.cu_seqlens_fo = None
        self.cu_seqlens_of = None
        self.max_seqlen_in_batch_o = None
        self.max_seqlen_in_batch_f = None
        self.max_seqlen_in_batch_fo = None
        self.max_seqlen_in_batch_of = None

    def _init_flash_attn(self, x: torch.Tensor, n_obs: int, n_feat: int):
        """Initialize GPU version with flash attention."""
        self.cpu_mode = False

        # Original flash attention initialization logic
        x_o, self.indices_o, self.cu_seqlens_o, self.max_seqlen_in_batch_o, *_ = unpad_input(x, ~self.padding_mask)

        self.feature_mask_big = einops.repeat(self.feature_mask, 'b f -> b s f', s=n_obs)
        self.feature_mask_big, _, _, _, *_ = unpad_input(self.feature_mask_big, ~self.padding_mask)
        x_of, self.indices_of, self.cu_seqlens_of, self.max_seqlen_in_batch_of, *_ = unpad_input(x_o, ~self.feature_mask_big)

        x_rearranged = einx.rearrange('b s f d -> b f s d', x)
        x_f, self.indices_f, self.cu_seqlens_f, self.max_seqlen_in_batch_f, *_ = unpad_input(x_rearranged, ~self.feature_mask)

        self.padding_mask_big = einops.repeat(self.padding_mask, 'b s -> b f s', f=n_feat)
        self.padding_mask_big, _, _, _, *_ = unpad_input(self.padding_mask_big, ~self.feature_mask)
        x_fo, self.indices_fo, self.cu_seqlens_fo, self.max_seqlen_in_batch_fo, *_ = unpad_input(x_f, ~self.padding_mask_big)

        self.batch_size_f = x_f.shape[0]
        self.batch_size_o = x_o.shape[0]

        t = torch.arange(self.indices_fo.shape[0]).unsqueeze(1).to(x.device)
        self.obs_to_feat_indices = self.base_to_feat(self.obs_to_base(t)).squeeze(1)
        self.feat_to_obs_indices = self.base_to_obs(self.feat_to_base(t)).squeeze(1)

    def base_to_obs(self, x: torch.Tensor) -> torch.Tensor:
        if hasattr(self, 'cpu_mode') and self.cpu_mode:
            # CPU fallback: reshape for standard attention
            # Convert from (b, s, f, d) to (b*s, f*d) or similar flattened format
            b, s, f, d = x.shape
            return x.view(b * s, f * d)

        # GPU path with flash attention
        x = einx.rearrange('b s f d -> b f s d', x)
        x, _, _, _, *_ = unpad_input(x, ~self.feature_mask)
        x, _, _, _, *_ = unpad_input(x, ~self.padding_mask_big)
        return x

    def base_to_feat(self, x: torch.Tensor) -> torch.Tensor:
        if hasattr(self, 'cpu_mode') and self.cpu_mode:
            # CPU fallback: reshape for standard attention
            # Convert from (b, s, f, d) to (b*f, s*d) or similar flattened format
            b, s, f, d = x.shape
            return x.view(b * f, s * d)

        # GPU path with flash attention
        x, _, _, _, *_ = unpad_input(x, ~self.padding_mask)
        x, _, _, _, *_ = unpad_input(x, ~self.feature_mask_big)
        return x

    def obs_to_base(self, x: torch.Tensor) -> torch.Tensor:
        if hasattr(self, 'cpu_mode') and self.cpu_mode:
            # CPU fallback: reshape back to base format
            # This is the inverse of base_to_obs
            total_elements = x.numel()
            expected_d = self.original_shape[-1]  # last dimension
            b, s, f = self.batch_size, self.n_obs, self.n_feat
            return x.view(b, s, f, expected_d)

        # GPU path with flash attention
        x = pad_input(x, self.indices_fo, self.batch_size_f, self.max_seqlen_in_batch_fo)
        x = pad_input(x, self.indices_f, self.batch_size, self.max_seqlen_in_batch_f)
        x = einx.rearrange('b f s d -> b s f d', x)
        return x

    def feat_to_base(self, x: torch.Tensor) -> torch.Tensor:
        if hasattr(self, 'cpu_mode') and self.cpu_mode:
            # CPU fallback: reshape back to base format
            # This is the inverse of base_to_feat
            total_elements = x.numel()
            expected_d = self.original_shape[-1]  # last dimension
            b, s, f = self.batch_size, self.n_obs, self.n_feat
            return x.view(b, s, f, expected_d)

        # GPU path with flash attention
        x = pad_input(x, self.indices_of, self.batch_size_o, self.max_seqlen_in_batch_of)
        x = pad_input(x, self.indices_o, self.batch_size, self.max_seqlen_in_batch_o)
        return x

    def obs_to_feat(self, x: torch.Tensor) -> torch.Tensor:
        if hasattr(self, 'cpu_mode') and self.cpu_mode:
            # CPU fallback: simple pass-through or basic reshaping
            return x

        # GPU path with flash attention
        x = x[self.obs_to_feat_indices]
        return x

    def feat_to_obs(self, x: torch.Tensor) -> torch.Tensor:
        if hasattr(self, 'cpu_mode') and self.cpu_mode:
            # CPU fallback: simple pass-through or basic reshaping
            return x

        # GPU path with flash attention
        x = x[self.feat_to_obs_indices]
        return x


class Layer(torch.nn.Module):

    def __init__(self, dim: int, n_heads: int, use_flash_attn: bool) -> None:

        super().__init__()

        self.layer_norm1 = nn.LayerNorm(dim)
        self.attention1 = MultiheadAttention(dim, n_heads, use_flash_attn)
        self.layer_norm2 = nn.LayerNorm(dim)
        self.linear1 = nn.Linear(dim, dim*4, bias=True)
        self.linear2 = nn.Linear(dim*4, dim, bias=True)

        self.layer_norm3 = nn.LayerNorm(dim)
        self.attention2 = MultiheadAttention(dim, n_heads, use_flash_attn)
        self.layer_norm4 = nn.LayerNorm(dim)
        self.linear3 = nn.Linear(dim, dim*4, bias=True)
        self.linear4 = nn.Linear(dim*4, dim, bias=True)

        self.use_flash_attn = use_flash_attn


    def forward(
            self,
            support: torch.Tensor,
            query__: torch.Tensor,
            padder_support: Optional[Padder],
            padder_query__: Optional[Padder],
            batch_size: Optional[int] = None,
            padding_obs_support: Optional[torch.Tensor] = None,
            padding_obs_query__: Optional[torch.Tensor] = None,
            padding_features: Optional[torch.Tensor] = None,
        ) -> tuple[torch.Tensor, torch.Tensor]:

        """
        Input:
        support in 'obs' format
        query__ in 'obs' format

        Output:
        support in 'obs' format
        query__ in 'obs' format
        """

        if self.use_flash_attn and padder_support is not None and padder_query__ is not None:
            support_residual = support
            query___residual = query__

            support = self.layer_norm1(support)
            query__ = self.layer_norm1(query__)

            # attention across rows
            support_att = self.attention1(
                support, support, support,
                cu_seqlens_q = padder_support.cu_seqlens_fo, max_seqlen_q = padder_support.max_seqlen_in_batch_fo,
                cu_seqlens_k = padder_support.cu_seqlens_fo, max_seqlen_k = padder_support.max_seqlen_in_batch_fo
            )
            query___att = self.attention1(
                query__, support, support,
                cu_seqlens_q = padder_query__.cu_seqlens_fo, max_seqlen_q = padder_query__.max_seqlen_in_batch_fo,
                cu_seqlens_k = padder_support.cu_seqlens_fo, max_seqlen_k = padder_support.max_seqlen_in_batch_fo
            )

            support = support_residual + support_att
            query__ = query___residual + query___att

            support_residual = support
            query___residual = query__

            support = self.layer_norm2(support)
            query__ = self.layer_norm2(query__)

            support = self.linear1(support)
            query__ = self.linear1(query__)

            support = torch.nn.functional.gelu(support)
            query__ = torch.nn.functional.gelu(query__)

            support = self.linear2(support)
            query__ = self.linear2(query__)

            support = support_residual + support
            query__ = query___residual + query__

            support = padder_support.obs_to_feat(support)
            query__ = padder_query__.obs_to_feat(query__)

            support_residual = support
            query___residual = query__

            support = self.layer_norm3(support)
            query__ = self.layer_norm3(query__)

            # attention across features
            support = self.attention2(
                support, support, support,
                cu_seqlens_q = padder_support.cu_seqlens_of, max_seqlen_q = padder_support.max_seqlen_in_batch_of,
                cu_seqlens_k = padder_support.cu_seqlens_of, max_seqlen_k = padder_support.max_seqlen_in_batch_of
            )
            query__ = self.attention2(
                query__, query__, query__,
                cu_seqlens_q = padder_query__.cu_seqlens_of, max_seqlen_q = padder_query__.max_seqlen_in_batch_of,
                cu_seqlens_k = padder_query__.cu_seqlens_of, max_seqlen_k = padder_query__.max_seqlen_in_batch_of
            )

            support = support_residual + support
            query__ = query___residual + query__

            support_residual = support
            query___residual = query__

            support = self.layer_norm4(support)
            query__ = self.layer_norm4(query__)

            support = self.linear3(support)
            query__ = self.linear3(query__)

            support = torch.nn.functional.gelu(support)
            query__ = torch.nn.functional.gelu(query__)

            support = self.linear4(support)
            query__ = self.linear4(query__)

            support = support_residual + support
            query__ = query___residual + query__

            support = padder_support.feat_to_obs(support)
            query__ = padder_query__.feat_to_obs(query__)

            return support, query__
        else:
            # CPU/Standard attention path - ensure it matches the GPU logic exactly
            # Input format: (b, s, f+1, d) where f+1 includes the y embedding
            batch_size_actual, n_obs_support, n_feat_plus_one, dim = support.shape
            _, n_obs_query, _, _ = query__.shape

            if batch_size is None:
                batch_size = batch_size_actual

            # First attention block - attention across observations (rows)
            support_residual = support
            query___residual = query__

            support = self.layer_norm1(support)
            query__ = self.layer_norm1(query__)

            # Reshape for row attention: (b, s, f+1, d) -> (b*(f+1), s, d)
            support_flat = einops.rearrange(support, 'b s f d -> (b f) s d')
            query___flat = einops.rearrange(query__, 'b s f d -> (b f) s d')

            # attention across observations
            support_att_flat = self.attention1(support_flat, support_flat, support_flat)
            query___att_flat = self.attention1(query___flat, support_flat, support_flat)

            # Reshape back: (b*(f+1), s, d) -> (b, s, f+1, d)
            support_att = einops.rearrange(support_att_flat, '(b f) s d -> b s f d', b=batch_size)
            query___att = einops.rearrange(query___att_flat, '(b f) s d -> b s f d', b=batch_size)

            support = support_residual + support_att
            query__ = query___residual + query___att

            # First MLP block
            support_residual = support
            query___residual = query__

            support = self.layer_norm2(support)
            query__ = self.layer_norm2(query__)

            support = self.linear1(support)
            query__ = self.linear1(query__)

            support = torch.nn.functional.gelu(support)
            query__ = torch.nn.functional.gelu(query__)

            support = self.linear2(support)
            query__ = self.linear2(query__)

            support = support_residual + support
            query__ = query___residual + query__

            # Second attention block - attention across features
            support_residual = support
            query___residual = query__

            support = self.layer_norm3(support)
            query__ = self.layer_norm3(query__)

            # Reshape for feature attention: (b, s, f+1, d) -> (b*s, f+1, d)
            support_feat = einops.rearrange(support, 'b s f d -> (b s) f d')
            query___feat = einops.rearrange(query__, 'b s f d -> (b s) f d')

            # attention across features
            support_feat_att = self.attention2(support_feat, support_feat, support_feat)
            query___feat_att = self.attention2(query___feat, query___feat, query___feat)

            # Reshape back: (b*s, f+1, d) -> (b, s, f+1, d)
            support_feat_att = einops.rearrange(support_feat_att, '(b s) f d -> b s f d', b=batch_size)
            query___feat_att = einops.rearrange(query___feat_att, '(b s) f d -> b s f d', b=batch_size)

            support = support_residual + support_feat_att
            query__ = query___residual + query___feat_att

            # Second MLP block
            support_residual = support
            query___residual = query__

            support = self.layer_norm4(support)
            query__ = self.layer_norm4(query__)

            support = self.linear3(support)
            query__ = self.linear3(query__)

            support = torch.nn.functional.gelu(support)
            query__ = torch.nn.functional.gelu(query__)

            support = self.linear4(support)
            query__ = self.linear4(query__)

            support = support_residual + support
            query__ = query___residual + query__

            return support, query__


class MultiheadAttention(torch.nn.Module):

    def __init__(self, dim: int, n_heads: int, use_flash_attn: bool) -> None:

        super().__init__()

        self.use_flash_attn = use_flash_attn
        self.dim = dim
        self.n_heads = n_heads

        self.q = nn.Linear(dim, dim, bias=True)
        self.k = nn.Linear(dim, dim, bias=True)
        self.v = nn.Linear(dim, dim, bias=True)
        self.o = nn.Linear(dim, dim, bias=True)


    def forward(
            self,
            query: torch.Tensor,
            key: torch.Tensor,
            value: torch.Tensor,
            cu_seqlens_q: Optional[torch.Tensor] = None,
            cu_seqlens_k: Optional[torch.Tensor] = None,
            max_seqlen_q: Optional[int] = None,
            max_seqlen_k: Optional[int] = None,
        ) -> torch.Tensor:
        """
        b = batch size
        s = number of observations
        f = number of features
        t = flashattention-compressed sequences of (batch, observations, features)
        h = heads
        d = dimension of embedding

        input: (bsf, d) for flash attention or (b, s, d) for standard attention
        output: (bsf, d) for flash attention or (b, s, d) for standard attention
        """

        q = self.q(query)
        k = self.k(key)
        v = self.v(value)

        if self.use_flash_attn and cu_seqlens_q is not None:
            q = einops.rearrange(q, 't (h d) -> t h d', h=self.n_heads) # (tokens, heads, dim), tokens is b*n*f w/o pad
            k = einops.rearrange(k, 't (h d) -> t h d', h=self.n_heads)
            v = einops.rearrange(v, 't (h d) -> t h d', h=self.n_heads)

            output = flash_attn_varlen_func(
                q = q,
                k = k,
                v = v,
                cu_seqlens_q = cu_seqlens_q, # num_seq+1, either b*n (w/o pad)+1, or b*f (w/o pad)+1
                cu_seqlens_k = cu_seqlens_k,
                max_seqlen_q = max_seqlen_q, # max sequence length, either n or f
                max_seqlen_k = max_seqlen_k,
                deterministic=True,
            )

            output = einops.rearrange(output, 't h d -> t (h d)')
        else:
            # Standard scaled dot-product attention for CPU
            q = einops.rearrange(q, 'b t (h d) -> b h t d', h=self.n_heads)
            k = einops.rearrange(k, 'b t (h d) -> b h t d', h=self.n_heads)
            v = einops.rearrange(v, 'b t (h d) -> b h t d', h=self.n_heads)

            output = F.scaled_dot_product_attention(q, k, v)
            output = einops.rearrange(output, 'b h t d -> b t (h d)')

        output = self.o(output)

        return output


import numpy as np

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import FunctionTransformer
from sklearn.preprocessing import QuantileTransformer

from .enums import Task


def create_y_transformer(y_train: np.ndarray, task: Task) -> TransformerMixin:
    # The y_transformer transforms the target variable to a normal distribution
    # This should be used for the y variable when training a regression model,
    # but when testing the model, we want to inverse transform the predictions

    if task == Task.REGRESSION:
        y_transformer = QuantileTransformer1D(output_distribution="uniform")
        y_transformer.fit(y_train)
        return y_transformer
    else:
        # Identity
        return FunctionTransformer()


class QuantileTransformer1D(BaseEstimator, TransformerMixin):

    def __init__(self, output_distribution="normal") -> None:
        self.quantile_transformer = QuantileTransformer(output_distribution=output_distribution)

    def fit(self, x: np.ndarray):
        self.quantile_transformer.fit(x[:, None])
        return self
    
    def transform(self, x: np.ndarray):
        return self.quantile_transformer.transform(x[:, None])[:, 0]
    
    def inverse_transform(self, x: np.ndarray):
        return self.quantile_transformer.inverse_transform(x[:, None])[:, 0]


from __future__ import annotations

import logging
import os
from typing import List, Optional

import pandas as pd

from autogluon.common.utils.resource_utils import ResourceManager
from autogluon.core.models import AbstractModel
from autogluon.features.generators import LabelEncoderFeatureGenerator
from autogluon.tabular import __version__

logger = logging.getLogger(__name__)


class MitraModel(AbstractModel):
    """
    Mitra is a tabular foundation model pre-trained purely on synthetic data with the goal
    of optimizing fine-tuning performance over in-context learning performance.
    Mitra was developed by the AutoGluon team @ AWS AI.

    Mitra's default hyperparameters outperforms all methods for small datasets on TabArena-v0.1 (excluding ensembling): https://tabarena.ai

    Authors: Xiyuan Zhang, Danielle C. Maddix, Junming Yin, Nick Erickson, Abdul Fatir Ansari, Boran Han, Shuai Zhang, Leman Akoglu, Christos Faloutsos, Michael W. Mahoney, Cuixiong Hu, Huzefa Rangwala, George Karypis, Bernie Wang
    Blog Post: https://www.amazon.science/blog/mitra-mixed-synthetic-priors-for-enhancing-tabular-foundation-models
    License: Apache-2.0

    .. versionadded:: 1.4.0
    """
    ag_key = "MITRA"
    ag_name = "Mitra"
    weights_file_name = "model.pt"
    ag_priority = 55

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._weights_saved = False
        self._feature_generator = None

    @staticmethod
    def _get_default_device():
        """Get the best available device for the current system."""
        if ResourceManager.get_gpu_count_torch(cuda_only=True) > 0:
            logger.log(15, "Using CUDA GPU")
            return "cuda"
        else:
            return "cpu"

    def get_model_cls(self):
        if self.problem_type in ["binary", "multiclass"]:
            from .sklearn_interface import MitraClassifier

            model_cls = MitraClassifier
        elif self.problem_type == "regression":
            from .sklearn_interface import MitraRegressor

            model_cls = MitraRegressor
        else:
            raise AssertionError(f"Unsupported problem_type: {self.problem_type}")
        return model_cls

    def _preprocess(self, X: pd.DataFrame, is_train: bool = False, **kwargs) -> pd.DataFrame:
        X = super()._preprocess(X, **kwargs)

        if is_train:
            # X will be the training data.
            self._feature_generator = LabelEncoderFeatureGenerator(verbosity=0)
            self._feature_generator.fit(X=X)

        # This converts categorical features to numeric via stateful label encoding.
        if self._feature_generator.features_in:
            X = X.copy()
            X[self._feature_generator.features_in] = self._feature_generator.transform(
                X=X
            )

        return X

    def _get_random_seed_from_hyperparameters(self, hyperparameters: dict) -> int | None | str:
        return hyperparameters.get("seed", "N/A")

    def _fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        X_val: pd.DataFrame = None,
        y_val: pd.Series = None,
        time_limit: float = None,
        num_cpus: int = 1,
        num_gpus: float = 0,
        verbosity: int = 2,
        **kwargs,
    ):
        # TODO: Reset the number of threads based on the specified num_cpus
        need_to_reset_torch_threads = False
        torch_threads_og = None

        try:
            model_cls = self.get_model_cls()
            import torch
        except ImportError as err:
            logger.log(
                40,
                f"\tFailed to import Mitra! To use the Mitra model, "
                f"do: `pip install autogluon.tabular[mitra]=={__version__}`.",
            )
            raise err

        if num_cpus is not None and isinstance(num_cpus, (int, float)):
            torch_threads_og = torch.get_num_threads()
            if torch_threads_og != num_cpus:
                # reset torch threads back to original value after fit
                torch.set_num_threads(num_cpus)
                need_to_reset_torch_threads = True

        hyp = self._get_model_params()

        hf_cls_model = hyp.pop("hf_cls_model", None)
        hf_reg_model = hyp.pop("hf_reg_model", None)
        if self.problem_type in ["binary", "multiclass"]:
            hf_model = hf_cls_model
        elif self.problem_type == "regression":
            hf_model = hf_reg_model
        else:
            raise AssertionError(f"Unsupported problem_type: {self.problem_type}")
        if hf_model is None:
            hf_model = hyp.pop("hf_general_model", None)
        if hf_model is None:
            hf_model = hyp.pop("hf_model", None)
        if hf_model is not None:
            logger.log(30, f"\tCustom hf_model specified: {hf_model}")
            hyp["hf_model"] = hf_model

        if hyp.get("device", None) is None:
            if num_gpus == 0:
                hyp["device"] = "cpu"
            else:
                hyp["device"] = self._get_default_device()

        if hyp["device"] == "cpu" and hyp.get("fine_tune", True):
            logger.log(
                30,
                f"\tWarning: Attempting to fine-tune Mitra on CPU. This will be very slow. "
                f"We strongly recommend using a GPU instance to fine-tune Mitra."
            )

        if "state_dict_classification" in hyp:
            state_dict_classification = hyp.pop("state_dict_classification")
            if self.problem_type in ["binary", "multiclass"]:
                hyp["state_dict"] = state_dict_classification
        if "state_dict_regression" in hyp:
            state_dict_regression = hyp.pop("state_dict_regression")
            if self.problem_type in ["regression"]:
                hyp["state_dict"] = state_dict_regression

        if "verbose" not in hyp:
            hyp["verbose"] = verbosity >= 3

        self.model = model_cls(
            seed=self.random_seed,
            **hyp,
        )

        X = self.preprocess(X, is_train=True)
        if X_val is not None:
            X_val = self.preprocess(X_val)

        self.model = self.model.fit(
            X=X,
            y=y,
            X_val=X_val,
            y_val=y_val,
            time_limit=time_limit,
        )

        if need_to_reset_torch_threads:
            torch.set_num_threads(torch_threads_og)

    def _set_default_params(self):
        default_params = {
            "n_estimators": 1,
        }
        for param, val in default_params.items():
            self._set_default_param_value(param, val)

    def _get_default_auxiliary_params(self) -> dict:
        default_auxiliary_params = super()._get_default_auxiliary_params()
        default_auxiliary_params.update(
            {
                "max_rows": 10000,
                "max_features": 500,
                "max_classes": 10,
            }
        )
        return default_auxiliary_params

    @property
    def weights_path(self) -> str:
        return os.path.join(self.path, self.weights_file_name)

    def save(self, path: str = None, verbose=True) -> str:
        _model_weights_list = None
        if self.model is not None:
            _model_weights_list = []
            for i in range(len(self.model.trainers)):
                _model_weights_list.append(self.model.trainers[i].model)
                self.model.trainers[i].checkpoint = None
                self.model.trainers[i].model = None
                self.model.trainers[i].optimizer = None
                self.model.trainers[i].scheduler_warmup = None
                self.model.trainers[i].scheduler_reduce_on_plateau = None
            self._weights_saved = True
        path = super().save(path=path, verbose=verbose)
        if _model_weights_list is not None:
            import torch

            os.makedirs(self.path, exist_ok=True)
            torch.save(_model_weights_list, self.weights_path)
            for i in range(len(self.model.trainers)):
                self.model.trainers[i].model = _model_weights_list[i]
        return path

    @classmethod
    def load(cls, path: str, reset_paths=False, verbose=True):
        model: MitraModel = super().load(path=path, reset_paths=reset_paths, verbose=verbose)

        if model._weights_saved:
            import torch

            model_weights_list = torch.load(model.weights_path, weights_only=False)  # nosec B614
            for i in range(len(model.model.trainers)):
                model.model.trainers[i].model = model_weights_list[i]
            model._weights_saved = False
        return model

    @classmethod
    def download_weights(cls, repo_id: str):
        """
        Download weights for Mitra from HuggingFace from `repo_id`.
        Requires an internet connection.
        """
        from huggingface_hub import hf_hub_download
        hf_hub_download(repo_id=repo_id, filename="config.json")
        hf_hub_download(repo_id=repo_id, filename="model.safetensors")

    @classmethod
    def download_default_weights(cls):
        """
        Download default weights for Mitra from HuggingFace.
        Includes both classifier and regressor weights.

        This is useful to call when building a docker image to avoid having to download Mitra weights for each instance.
        This is also useful for benchmarking as a first sanity check
        to avoid HuggingFace potentially blocking the download.

        Requires an internet connection.
        """
        cls.download_weights(repo_id="autogluon/mitra-classifier")
        cls.download_weights(repo_id="autogluon/mitra-regressor")

    @classmethod
    def supported_problem_types(cls) -> Optional[List[str]]:
        return ["binary", "multiclass", "regression"]

    @classmethod
    def _get_default_ag_args_ensemble(cls, **kwargs) -> dict:
        default_ag_args_ensemble = super()._get_default_ag_args_ensemble(**kwargs)
        # FIXME: Test if it works with parallel, need to enable n_cpus support
        extra_ag_args_ensemble = {
            "fold_fitting_strategy": "sequential_local",  # FIXME: Comment out after debugging for large speedup
        }
        default_ag_args_ensemble.update(extra_ag_args_ensemble)
        return default_ag_args_ensemble

    def _get_default_resources(self) -> tuple[int, int]:
        # Use only physical cores for better performance based on benchmarks
        num_cpus = ResourceManager.get_cpu_count(only_physical_cores=True)

        num_gpus = min(1, ResourceManager.get_gpu_count_torch(cuda_only=True))

        return num_cpus, num_gpus

    def _estimate_memory_usage(self, X: pd.DataFrame, **kwargs) -> int:
        return self.estimate_memory_usage_static(
            X=X, problem_type=self.problem_type, num_classes=self.num_classes, **kwargs
        )

    @classmethod
    def _estimate_memory_usage_static(
        cls,
        *,
        X: pd.DataFrame,
        **kwargs,
    ) -> int:
        # Multiply by 0.9 as currently this is overly safe
        return int(0.9 * max(
            cls._estimate_memory_usage_static_cpu_icl(X=X, **kwargs),
            cls._estimate_memory_usage_static_cpu_ft_icl(X=X, **kwargs),
            cls._estimate_memory_usage_static_gpu_cpu(X=X, **kwargs),
            cls._estimate_memory_usage_static_gpu_gpu(X=X, **kwargs),
        ))

    @classmethod
    def _estimate_memory_usage_static_cpu_icl(
        cls,
        *,
        X: pd.DataFrame,
        **kwargs,
    ) -> int:
        rows, features = X.shape[0], X.shape[1]

        # For very small datasets, use a more conservative estimate
        if rows * features < 100:  # Small dataset threshold
            # Use a simpler linear formula for small datasets
            cpu_memory_kb = 1.3 * (100 * rows * features + 1000000)  # 1GB base + linear scaling
        else:
            # Original formula for larger datasets
            cpu_memory_kb = 1.3 * (
                0.001748 * (rows**2) * features + 0.001206 * rows * (features**2) + 10.3482 * rows * features + 6409698
            )
        return int(cpu_memory_kb * 1e3)

    @classmethod
    def _estimate_memory_usage_static_cpu_ft_icl(
        cls,
        *,
        X: pd.DataFrame,
        **kwargs,
    ) -> int:
        rows, features = X.shape[0], X.shape[1]

        # For very small datasets, use a more conservative estimate
        if rows * features < 100:  # Small dataset threshold
            # Use a simpler linear formula for small datasets
            cpu_memory_kb = 1.3 * (200 * rows * features + 2000000)  # 2GB base + linear scaling
        else:
            # Original formula for larger datasets
            cpu_memory_kb = 1.3 * (
                0.001 * (rows**2) * features + 0.004541 * rows * (features**2) + 46.2974 * rows * features + 5605681
            )
        return int(cpu_memory_kb * 1e3)

    @classmethod
    def _estimate_memory_usage_static_gpu_cpu(
        cls,
        *,
        X: pd.DataFrame,
        **kwargs,
    ) -> int:
        rows, features = X.shape[0], X.shape[1]

        # For very small datasets, use a more conservative estimate
        if rows * features < 100:  # Small dataset threshold
            return int(2.5 * 1e9)  # 2.5GB for small datasets
        else:
            return int(5 * 1e9)  # 5GB for larger datasets

    @classmethod
    def _estimate_memory_usage_static_gpu_gpu(
        cls,
        *,
        X: pd.DataFrame,
        **kwargs,
    ) -> int:
        rows, features = X.shape[0], X.shape[1]

        # For very small datasets, use a more conservative estimate
        if rows * features < 100:  # Small dataset threshold
            # Use a simpler linear formula for small datasets
            gpu_memory_mb = 1.3 * (10 * rows * features + 2000)  # 2GB base + linear scaling
        else:
            # Original formula for larger datasets
            gpu_memory_mb = 1.3 * (0.05676 * rows * features + 3901)
        return int(gpu_memory_mb * 1e6)

    @classmethod
    def _class_tags(cls) -> dict:
        return {
            "can_estimate_memory_usage_static": True,
        }

    def _more_tags(self) -> dict:
        tags = {"can_refit_full": True}
        return tags



from __future__ import annotations

import logging
import os
from pathlib import Path
import time

import numpy as np
import pandas as pd

from autogluon.common.utils.pandas_utils import get_approximate_df_mem_usage
from autogluon.common.utils.resource_utils import ResourceManager
from autogluon.common.utils.try_import import try_import_torch
from autogluon.core.constants import BINARY, MULTICLASS, REGRESSION, QUANTILE
from autogluon.core.models import AbstractModel
from autogluon.core.utils import generate_train_test_split
from autogluon.core.utils.exceptions import TimeLimitExceeded
from autogluon.features.generators import LabelEncoderFeatureGenerator

logger = logging.getLogger(__name__)


class TabPFNMixModel(AbstractModel):
    """
    [Experimental model] Can be changed/removed without warning in future releases.

    TabPFNMix is based off of the TabPFN and TabForestPFN models.

    We recommend using Mitra instead, as it is an improved version of TabPFNMix.

    It is a tabular transformer model pre-trained on purely synthetic data.

    It currently has several limitations:
    1. Does not support regression
    2. Does not support >10 classes
    3. Does not support GPU

    For more information, refer to the `./_internals/README.md` file.

    .. versionadded:: 1.2.0
    """
    ag_key = "TABPFNMIX"
    ag_name = "TabPFNMix"
    ag_priority = 45

    weights_file_name = "model.pt"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._feature_generator = None
        self._weights_saved = False

    def _get_model_type(self):
        from ._internal.tabpfnmix_classifier import TabPFNMixClassifier
        from ._internal.tabpfnmix_regressor import TabPFNMixRegressor
        if self.problem_type in ['binary', 'multiclass']:
            model_cls = TabPFNMixClassifier
        elif self.problem_type in ['regression']:
            model_cls = TabPFNMixRegressor
        else:
            raise AssertionError(f"TabPFN does not support problem_type='{self.problem_type}'")
        return model_cls

    def _set_default_params(self):
        """Specifies hyperparameter values to use by default"""
        default_params = {
            # most important hyperparameters. Only set `n_estimators>1` if `max_epochs>1`, else there will be no benefit.
            # model_path,  # most important, defines huggingface model path
            "model_path_classifier": "autogluon/tabpfn-mix-1.0-classifier",  # if specified, overrides model_path for classification problems, set to None to ignore.
            "model_path_regressor": "autogluon/tabpfn-mix-1.0-regressor",  # if specified, overrides model_path for regression problems, set to None to ignore.
            # weights_path,  # most important, defines weights location (overrides huggingface weights if specified)
            # weights_path_classifier,  # if specified, overrides weights_path for classification problems
            # weights_path_regressor,  # if specified, overrides weights_path for regression problems
            "n_ensembles": 1,  # FIXME: RENAME: n_estimators
            "max_epochs": 0,  # fine-tuning epochs. Will do pure in-context learning if 0.

            # next most important hyperparameters
            "lr": 1.0e-05,
            "max_samples_query": 1024,  # larger = slower but better quality on datasets with at least this many validation samples
            "max_samples_support": 8196,  # larger = slower but better quality on datasets with at least this many training samples

            # other hyperparameters
            "early_stopping_patience": 40,  # TODO: Figure out optimal value
            "linear_attention": True,
            "lr_scheduler": False,
            "lr_scheduler_patience": 30,
            "optimizer": "adamw",
            "use_feature_count_scaling": True,
            "use_quantile_transformer": True,
            "weight_decay": 0,

            # architecture hyperparameters, recommended to keep as default unless using a custom pre-trained backbone
            "n_classes": 10,
            "n_features": 100,
            "n_heads": 4,
            "n_layers": 12,
            "attn_dropout": 0.0,
            "dim": 512,
            "y_as_float_embedding": True,

            # utility parameters, recommended to keep as default
            "split_val": False,
            "use_best_epoch": True,
        }
        for param, val in default_params.items():
            self._set_default_param_value(param, val)

    # FIXME: Make X_val, y_val = None work well, currently it uses X and y as validation, should instead skip validation entirely
    # FIXME: Use `params_trained` for refit optimal epochs
    # FIXME: Handle model weights download
    # FIXME: GPU support?
    # FIXME: Save model weights to file instead of pickling?
    def _fit(self, X: pd.DataFrame, y: pd.Series, X_val: pd.DataFrame = None, y_val: pd.Series = None, time_limit: float = None, num_cpus: int = 1, num_gpus: float = 0, **kwargs):
        time_start = time.time()
        try_import_torch()
        import torch
        from ._internal.config.config_run import ConfigRun

        ag_params = self._get_ag_params()
        max_classes = ag_params.get("max_classes")
        if max_classes is not None and self.num_classes is not None and self.num_classes > max_classes:
            # TODO: Move to earlier stage when problem_type is checked
            raise AssertionError(f"Max allowed classes for the model is {max_classes}, " f"but found {self.num_classes} classes.")

        params = self._get_model_params()
        sample_rows = ag_params.get("sample_rows", None)
        sample_rows_val = ag_params.get("sample_rows_val", None)
        max_rows = ag_params.get("max_rows", None)

        # TODO: Make max_rows generic
        if max_rows is not None and isinstance(max_rows, (int, float)) and len(X) > max_rows:
            raise AssertionError(f"Skipping model due to X having more rows than `ag.max_rows={max_rows}` (len(X)={len(X)})")

        # TODO: Make sample_rows generic
        if sample_rows is not None and isinstance(sample_rows, int) and len(X) > sample_rows:
            X, y = self._subsample_data(X=X, y=y, num_rows=sample_rows)

        # TODO: Make sample_rows generic
        if X_val is not None and y_val is not None and sample_rows_val is not None and isinstance(sample_rows_val, int) and len(X_val) > sample_rows_val:
            X_val, y_val = self._subsample_data(X=X_val, y=y_val, num_rows=sample_rows_val)

        from ._internal.core.enums import Task
        if self.problem_type in [REGRESSION, QUANTILE]:
            task = Task.REGRESSION
            n_classes = 0
        else:
            task = Task.CLASSIFICATION
            n_classes = self.num_classes

        if num_gpus > 0:
            device = "cuda:0"
        else:
            device = "cpu"

        model_path = None
        if task == Task.CLASSIFICATION:
            if "model_path_classifier" in params and params["model_path_classifier"] is not None:
                model_path = params["model_path_classifier"]
        elif task == Task.REGRESSION:
            if "model_path_regressor" in params and params["model_path_regressor"] is not None:
                model_path = params["model_path_regressor"]
        if model_path is None:
            model_path = params.get("model_path", None)

        weights_path = None
        if task == Task.CLASSIFICATION:
            if "weights_path_classifier" in params and params["weights_path_classifier"] is not None:
                weights_path = Path(params["weights_path_classifier"])
        elif task == Task.REGRESSION:
            if "weights_path_regressor" in params and params["weights_path_regressor"] is not None:
                weights_path = Path(params["weights_path_regressor"])
        if weights_path is None:
            if "weights_path" in params and params["weights_path"] is not None:
                weights_path = Path(params["weights_path"])

        if weights_path is None and model_path is None:
            logger.log(15, "\tNo model_path or weights_path specified, fitting model from random initialization...")
        elif weights_path is not None:
            logger.log(15, f'\tLoading pre-trained weights from file... (weights_path="{weights_path}")')

        cfg = ConfigRun(hyperparams=params, task=task, device=device, seed=self.random_seed)

        if cfg.hyperparams["max_epochs"] == 0 and cfg.hyperparams["n_ensembles"] != 1:
            logger.log(
                30,
                f"WARNING: max_epochs should be > 0 if n_ensembles > 1, otherwise there will be zero quality benefit with slower inference. "
                f"(max_epochs={cfg.hyperparams['max_epochs']}, n_ensembles={cfg.hyperparams['n_ensembles']})"
            )

        X = self.preprocess(X)
        y = y.values
        if X_val is not None:
            X_val = self.preprocess(X_val)
            y_val = y_val.values

        # FIXME: What if an exception occurs or timeout occurs after updating threads? Can we add logic to ensure torch num_threads are reset?
        need_to_reset_torch_threads = False
        torch_threads_og = None
        if num_cpus is not None and isinstance(num_cpus, (int, float)):
            torch_threads_og = torch.get_num_threads()
            if torch_threads_og != num_cpus:
                # reset torch threads back to original value after fit
                torch.set_num_threads(num_cpus)
                need_to_reset_torch_threads = True

        model_cls = self._get_model_type()

        if time_limit is not None:
            time_cur = time.time()
            time_left = time_limit - (time_cur - time_start)
            if time_left <= 0:
                raise TimeLimitExceeded(f"No time remaining to fit model (time_limit={time_limit:.2f}s, time_left={time_left:.2f}s)")
            time_limit = time_left

        self.model = model_cls(
            cfg=cfg,
            n_classes=n_classes,
            split_val=params["split_val"],
            model_path=model_path,
            weights_path=weights_path,
            stopping_metric=self.stopping_metric,
            use_best_epoch=params["use_best_epoch"],
        )
        self.model.fit(
            X=X,
            y=y,
            X_val=X_val,
            y_val=y_val,
            time_limit=time_limit,
        )

        # Ensure refit_full uses the same number of max_epochs as the original's best
        self.params_trained["max_epochs"] = self.model.trainer.best_epoch
        self.params_trained["ag.max_rows"] = None  # This ensures we don't raise an exception during refit_full

        # reduce memory and disk usage by 3x
        self.model.trainer.minimize_for_inference()

        if need_to_reset_torch_threads:
            torch.set_num_threads(torch_threads_og)

        return self

    # TODO: Make this generic by creating a generic `preprocess_train` and putting this logic prior to `_preprocess`.
    def _subsample_data(self, X: pd.DataFrame, y: pd.Series, num_rows: int) -> (pd.DataFrame, pd.Series):
        num_rows_to_drop = len(X) - num_rows
        X, _, y, _ = generate_train_test_split(
            X=X,
            y=y,
            problem_type=self.problem_type,
            test_size=num_rows_to_drop,
            random_state=self.random_seed,
            min_cls_count_train=1,
        )
        return X, y

    def _preprocess(self, X: pd.DataFrame, **kwargs) -> np.ndarray:
        """
        Converts categorical to label encoded integers
        Keeps missing values, as TabPFN automatically handles missing values internally.
        """
        X = super()._preprocess(X, **kwargs)
        if self._feature_generator is None:
            # FIXME: Check if this is needed, never actually tried removing it, copy pasted from TabPFNModel implementation in AG
            self._feature_generator = LabelEncoderFeatureGenerator(verbosity=0)
            self._feature_generator.fit(X=X)
        if self._feature_generator.features_in:
            X = X.copy()
            X[self._feature_generator.features_in] = self._feature_generator.transform(X=X)
        X = X.values.astype(np.float64)
        return X

    # FIXME: Switch to `torch.save(_model_weights.state_dict(), PATH)`, need to reinitialize the model though...
    def save(self, path: str = None, verbose=True) -> str:
        _model_weights = None
        if self.model is not None:
            _model_weights = self.model.trainer.model
            self.model.trainer.model = None
            self._weights_saved = True
        path = super().save(path=path, verbose=verbose)
        if _model_weights is not None:
            import torch
            os.makedirs(self.path, exist_ok=True)
            torch.save(_model_weights, self.weights_path)
            self.model.trainer.model = _model_weights
        return path

    # FIXME: Switch to `weights_only=True`, need to reinitialize the model though...
    @classmethod
    def load(cls, path: str, reset_paths=False, verbose=True):
        model: TabPFNMixModel = super().load(path=path, reset_paths=reset_paths, verbose=verbose)

        if model._weights_saved:
            import torch
            model.model.trainer.model = torch.load(model.weights_path, weights_only=False)  # nosec B614
            model._weights_saved = False
        return model

    @property
    def weights_path(self) -> str:
        return os.path.join(self.path, self.weights_file_name)

    @classmethod
    def supported_problem_types(cls) -> list[str] | None:
        return ["binary", "multiclass", "regression"]

    def _get_default_auxiliary_params(self) -> dict:
        default_auxiliary_params = super()._get_default_auxiliary_params()
        default_auxiliary_params.update(
            {
                "max_classes": 10,
            }
        )
        return default_auxiliary_params

    def _get_maximum_resources(self) -> dict[str, int | float]:
        # torch model trains slower when utilizing virtual cores and this issue scale up when the number of cpu cores increases
        return {"num_cpus": ResourceManager.get_cpu_count(only_physical_cores=True)}

    def _get_default_resources(self) -> tuple[int, float]:
        # only_physical_cores=True is faster in training
        num_cpus = ResourceManager.get_cpu_count(only_physical_cores=True)
        num_gpus = 0
        return num_cpus, num_gpus

    def _estimate_memory_usage(self, X: pd.DataFrame, **kwargs) -> int:
        hyperparameters = self._get_model_params()
        return self.estimate_memory_usage_static(X=X, problem_type=self.problem_type, num_classes=self.num_classes, hyperparameters=hyperparameters, **kwargs)

    def get_minimum_ideal_resources(self) -> dict[str, int | float]:
        return {"num_cpus": 4}

    @classmethod
    def _estimate_memory_usage_static(
        cls,
        *,
        X: pd.DataFrame,
        **kwargs,
    ) -> int:
        # TODO: This is wildly inaccurate, find a better estimation
        # TODO: Fitting 8 in parallel causes many OOM errors with 32 GB of memory on relatively small datasets, so each model is using over 4 GB of memory
        # TODO: Fitting 4 in parallel still causes many OOM errors with 32 GB of memory on relatively small datasets, so each model is using over 8 GB of memory
        #  The below logic returns a minimum of 8.8 GB, to avoid OOM errors
        data_mem_usage = 5 * get_approximate_df_mem_usage(X).sum()  # rough estimate
        model_size = 160*1e6  # model weights are ~160 MB  # TODO: Avoid hardcoding, we can derive from the model itself?
        model_mem_usage = model_size * 5  # Account for 1x copy being fit, 1x copy checkpointed, 2x for optimizer, and 1x for overhead
        model_fit_usage = model_size * 50  # TODO: This is a placeholder large value to try to avoid OOM errors
        mem_usage_estimate = data_mem_usage + model_mem_usage + model_fit_usage
        return mem_usage_estimate

    @classmethod
    def _class_tags(cls) -> dict:
        return {
            "can_estimate_memory_usage_static": True,
        }

    def _ag_params(self) -> set:
        return {"max_classes", "max_rows", "sample_rows", "sample_rows_val"}

    def _more_tags(self) -> dict:
        tags = {"can_refit_full": True}
        return tags


import copy

from autogluon.features.generators import (
    AutoMLInterpretablePipelineFeatureGenerator,
    AutoMLPipelineFeatureGenerator,
    IdentityFeatureGenerator,
)


def get_default_feature_generator(feature_generator, feature_metadata=None, init_kwargs=None):
    if init_kwargs is None:
        init_kwargs = dict()
    if feature_generator is None:
        feature_generator = IdentityFeatureGenerator()
    elif isinstance(feature_generator, str):
        if feature_generator == "auto":
            feature_generator = AutoMLPipelineFeatureGenerator(**init_kwargs)
        elif feature_generator == "interpretable":
            feature_generator = AutoMLInterpretablePipelineFeatureGenerator(**init_kwargs)
        else:
            raise ValueError(f"Unknown feature_generator preset: '{feature_generator}', valid presets: {['auto', 'interpretable']}")
    if feature_metadata is not None:
        if feature_generator.feature_metadata_in is None and not feature_generator.is_fit():
            feature_generator.feature_metadata_in = copy.deepcopy(feature_metadata)
        else:
            raise AssertionError("`feature_metadata_in` already exists in `feature_generator`.")
    return feature_generator
